"""
Definition of the loss function for the network
"""

import tensorflow as tf


def angle_diff(a, b):
    """
    Root Mean Square Error of the angle difference.
    The angle difference function takes into account the periodicity of angles
    """
    diff = tf.atan2(tf.sin(a - b), tf.cos(a - b))
    return tf.sqrt(tf.reduce_mean(tf.square(diff), axis=1))


def symmetric_angle_difference(y_true, y_pred):
    """
    We calculate the angle difference between the prediction and the two possible labels,
    and pick the minimum of the two,
    we average the result on the batch
    """

    dists = [angle_diff(y_pred, y_true[:, 0]), angle_diff(y_pred, y_true[:, 1])]
    mins = tf.reduce_min(dists, axis=0)
    loss = tf.reduce_mean(mins)
    return loss



from abc import ABC
from contextlib import AbstractContextManager
from typing import Sequence

import numpy as np

from wormpose import BaseFramesDataset


class BaseScoringDataManager(AbstractContextManager, Sequence, ABC):
    """
    Accessor for the data necessary to perform the scoring, including:
    - The image that needs to be scored
    - The image that will be used as a template for scoring, plus relevant infos: skeleton and measurements
    """

    pass


class ScoringDataManager(BaseScoringDataManager):
    """
    Implementation of the BaseScoringDataManager for scoring videos:
    it finds the closest template in time for a frame in a video
    """

    def __init__(self, frames_dataset: BaseFramesDataset, features, video_name: str):
        self.features = features
        self.video_name = video_name

        self._frames_dataset = frames_dataset
        self._frames_reader = None
        self._frames = None

    def __enter__(self):
        self._frames_reader = self._frames_dataset.open(self.video_name)
        self._frames = self._frames_reader.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._frames_reader.__exit__(exc_type, exc_value, traceback)
        self._frames_reader = None
        self._frames = None

    def __getitem__(self, frame_index: int):
        template_indexes = self.features.labelled_indexes
        # find the template image the most closely in time to the frame to score
        template_index = template_indexes[np.abs(template_indexes - frame_index).argmin()]
        return (
            self._frames[template_index],
            self.features.skeletons[template_index],
            self.features.measurements,
            self._frames[frame_index],
        )

    def __len__(self):
        return len(self._frames)



"""
This module exposes the gaussian mixture model of worm postures
from the resource file "postures_model.json.gz"
"""

import gzip
import json
import logging
import os
from typing import Generator

import numpy as np

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class PosturesModel(object):
    def __init__(self):

        model_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), os.pardir, "resources", "postures_model.json.gz",
        )

        with gzip.open(model_path, "rt") as f:
            shapes_model = json.load(f)

        self.means = np.array(shapes_model["means"])
        self.covariances = np.array(shapes_model["covariances"])
        self.weights = np.array(shapes_model["weights"])
        self.num_gaussians = self.means.shape[0]

    def generate(self) -> Generator:
        """
        Generates worm postures

        :return: a generator of numpy arrays of shape (N,) representing the centerline angles
        """
        while True:
            idx = np.random.choice(np.arange(self.num_gaussians), p=self.weights)
            random_global_angle = np.random.uniform(0, 2 * np.pi)

            samples = np.random.multivariate_normal(self.means[idx], self.covariances[idx], size=1)

            theta = samples[0] + random_global_angle

            yield theta



from typing import Optional

import numpy as np

from wormpose.pose.results_datatypes import BaseResults


def _resample(values: Optional[np.ndarray], timestamp: np.ndarray) -> Optional[np.ndarray]:
    if values is None:
        return values
    values = np.array(values)
    resampled = np.full((timestamp[-1] - timestamp[0] + 1,) + values[0].shape, np.nan, values.dtype)
    for i in range(len(values)):
        t = timestamp[i] - timestamp[0]
        resampled[t] = values[i]
    return resampled


def resample_results(results: BaseResults, timestamp: np.ndarray):
    results.scores = _resample(results.scores, timestamp)
    results.theta = _resample(results.theta, timestamp)
    results.skeletons = _resample(results.skeletons, timestamp)



"""
This module contains data structures representing tracking results
They all contain theta, scores and skeletons but we use several subclasses to specify
the current step of the tracking pipeline
"""
from typing import Optional

import numpy as np

from wormpose.pose.centerline import flip_theta_series


class BaseResults(object):
    def __init__(
        self,
        theta: Optional[np.ndarray] = None,
        skeletons: Optional[np.ndarray] = None,
        scores: Optional[np.ndarray] = None,
    ):
        self.theta = theta
        self.scores = scores
        self.skeletons = skeletons

    def __len__(self):
        return len(self.theta)


class OriginalResults(BaseResults):
    pass


class ShuffledResults(BaseResults):
    def __init__(self, random_theta: np.ndarray):
        shuffled_theta = np.stack([random_theta, flip_theta_series(random_theta)], axis=1)
        super().__init__(theta=shuffled_theta, scores=None, skeletons=None)



"""
This module assembles image data batches by preprocessing each image with a BaseFramePreprocessing object
It distributes the work to several processes and provides each batch as they are ready.
"""

import math
import os
import pickle
from multiprocessing import Process, Manager

import numpy as np

from wormpose import BaseFramePreprocessing
from wormpose.dataset import Dataset
from wormpose.images.real_dataset import RealDataset


class _ImagesBatch(object):
    def __init__(self, batch_size: int, image_shape):
        self.data = np.zeros(([batch_size, image_shape[0], image_shape[1], 1]), dtype=np.float32)
        self.index = 0

    def reset(self):
        self.index = 0

    def add(self, image: np.ndarray):
        self.data[self.index, :, :, 0] = image
        self.index += 1


def _assemble_images_batch(
    frame_preprocessing: BaseFramePreprocessing,
    data_reading_queue,
    results_queue,
    temp_dir: str,
    batch_size: int,
    image_shape,
):

    images_batch = _ImagesBatch(batch_size, image_shape)
    real_dataset = RealDataset(frame_preprocessing, image_shape)
    while True:
        queue_data = data_reading_queue.get()
        if queue_data is None:
            break

        data_filename, chunk_index = queue_data

        with open(data_filename, "rb") as f:
            images_batch.reset()
            while True:
                try:
                    raw_frame = pickle.load(f)
                except EOFError:
                    break
                cur_frame, _ = real_dataset.process_frame(raw_frame)
                images_batch.add(cur_frame)

        os.remove(data_filename)

        # normalize between  and 1
        images_data_batch = images_batch.data / 255.0

        image_filename = os.path.join(temp_dir, f"real_topredict_{chunk_index:09d}.npy")
        np.save(image_filename, images_data_batch)
        results_queue.put(image_filename)


class PredictDataGenerator(object):
    def __init__(
        self, dataset: Dataset, num_process: int, temp_dir: str, image_shape, batch_size: int,
    ):
        self.dataset = dataset
        self.num_process = num_process
        self.temp_dir = temp_dir
        self.image_shape = image_shape
        self.batch_size = batch_size

    def _read_data(self, data_reading_queue, num_frames: int, video_name: str):

        with self.dataset.frames_dataset.open(video_name) as frames:
            num_chunks = math.ceil(float(num_frames) / self.batch_size)
            for chunk_index in range(num_chunks):
                start = chunk_index * self.batch_size
                end = min(num_frames, start + self.batch_size)

                data_filename = os.path.join(self.temp_dir, f"data_to_predict_{chunk_index}.npy")
                with open(data_filename, "wb") as f:
                    for index in range(start, end):
                        pickle.dump(frames[index], f)

                data_reading_queue.put((data_filename, chunk_index))

        for _ in range(self.num_process):
            data_reading_queue.put(None)

    def run(self, video_name: str):

        manager = Manager()
        data_reading_queue = manager.Queue()
        results_queue = manager.Queue()

        workers = [
            Process(
                target=_assemble_images_batch,
                args=(
                    self.dataset.frame_preprocessing,
                    data_reading_queue,
                    results_queue,
                    self.temp_dir,
                    self.batch_size,
                    self.image_shape,
                ),
            )
            for _ in range(self.num_process)
        ]
        for w in workers:
            w.start()

        self._read_data(
            data_reading_queue=data_reading_queue,
            num_frames=self.dataset.num_frames(video_name),
            video_name=video_name,
        )

        for w in workers:
            w.join()

        results_files = []
        while not results_queue.empty():
            results_files.append(results_queue.get())
        results_files = sorted(results_files)

        for f in results_files:
            batch_data = np.load(f)
            yield batch_data
            os.remove(f)



#!/usr/bin/env python

"""
Calculates the image similarity on a random selection of labeled frames from a dataset.
"""

import logging
import os
import random
from argparse import Namespace
from typing import Tuple

import h5py
import numpy as np

from wormpose.commands import _log_parameters
from wormpose.config import default_paths
from wormpose.dataset import Dataset
from wormpose.dataset.loader import get_dataset_name
from wormpose.dataset.loader import load_dataset
from wormpose.images.scoring.centerline_accuracy_check import CenterlineAccuracyCheck
from wormpose.pose.centerline import skeletons_to_angles
from wormpose.dataset.loaders.resizer import add_resizing_arguments, ResizeOptions

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class _ScoresWriter(object):
    def __init__(self):
        self.all_scores = []

    def add(self, kwargs):
        self.all_scores.append(kwargs["score"])

    def write(self, results_file):
        with h5py.File(results_file, "a") as f:
            f.create_dataset("scores", data=self.all_scores)


class _ImagesAndScoresWriter(_ScoresWriter):
    def __init__(self):
        self.all_synth = []
        self.all_real = []
        super().__init__()

    def add(self, kwargs):
        centerline_accuracy: CenterlineAccuracyCheck = kwargs["centerline_accuracy"]
        self.all_real.append(np.array(centerline_accuracy.last_real_image))
        self.all_synth.append(np.array(centerline_accuracy.last_synth_image))
        super().add(kwargs)

    def write(self, results_file):
        with h5py.File(results_file, "a") as f:
            f.create_dataset("real_images", data=self.all_real)
            f.create_dataset("synth_images", data=self.all_synth)
        super().write(results_file)


class _Calibrator(object):
    def __init__(
        self, dataset: Dataset, results_dir: str, image_shape: Tuple[int, int], num_samples: int, theta_dims: int,
    ):
        self.dataset = dataset
        self.results_dir = results_dir
        self.image_shape = image_shape
        self.num_samples = num_samples
        self.theta_dims = theta_dims

    def __call__(self, video_name: str, writer: _ScoresWriter):
        """
        Evaluate image metric score on labelled frames.
        """
        features = self.dataset.features_dataset[video_name]
        labelled_thetas = skeletons_to_angles(features.skeletons, theta_dims=self.theta_dims)
        labelled_indexes = features.labelled_indexes

        centerline_accuracy = CenterlineAccuracyCheck(
            frame_preprocessing=self.dataset.frame_preprocessing, image_shape=self.image_shape,
        )

        with self.dataset.frames_dataset.open(video_name) as frames:
            frames_amount = min(self.num_samples, len(labelled_indexes))

            random_label_index = np.random.choice(labelled_indexes, frames_amount, replace=False)
            thetas = labelled_thetas[random_label_index]

            for theta, index in zip(thetas, random_label_index):
                cur_frame = frames[index]

                score, _ = centerline_accuracy(
                    theta=theta,
                    template_skeleton=features.skeletons[index],
                    template_measurements=features.measurements,
                    template_frame=cur_frame,
                    real_frame_orig=cur_frame,
                )
                writer.add(locals())

        results_file = os.path.join(self.results_dir, video_name + "_calibration.h5")
        if os.path.exists(results_file):
            os.remove(results_file)
        writer.write(results_file=results_file)

        logger.info(
            f"Evaluated known skeletons reconstruction for {video_name},"
            f" average score {np.mean(writer.all_scores):.4f}"
        )
        return results_file


def _parse_arguments(kwargs: dict):
    if kwargs.get("num_samples") is None:
        kwargs["num_samples"] = 500
    if kwargs.get("work_dir") is None:
        kwargs["work_dir"] = default_paths.WORK_DIR
    if kwargs.get("theta_dims") is None:
        kwargs["theta_dims"] = 100
    if kwargs.get("video_names") is None:
        kwargs["video_names"] = None
    if kwargs.get("save_images") is None:
        kwargs["save_images"] = False
    if kwargs.get("random_seed") is None:
        kwargs["random_seed"] = None
    kwargs["resize_options"] = ResizeOptions(**kwargs)

    _log_parameters(logger.info, kwargs)
    return Namespace(**kwargs)


def calibrate(dataset_loader: str, dataset_path: str, **kwargs):
    """
    Calculate the image score for a certain number of labelled frames in the dataset,
    this will give an indication on choosing the image similarity threshold when predicting all frames in the dataset.

    :param dataset_loader: Name of the dataset loader, for example "tierpsy"
    :param dataset_path: Root path of the dataset containing videos of worm
    """
    _log_parameters(logger.info, {"dataset_loader": dataset_loader, "dataset_path": dataset_path})
    args = _parse_arguments(kwargs)

    random.seed(args.random_seed)
    np.random.seed(args.random_seed)

    dataset_name = get_dataset_name(dataset_path)
    experiment_dir = os.path.join(args.work_dir, dataset_name)
    calibration_results_dir = os.path.join(experiment_dir, default_paths.CALIBRATION_RESULTS_DIR)
    os.makedirs(calibration_results_dir, exist_ok=True)

    dataset = load_dataset(
        dataset_loader, dataset_path, resize_options=args.resize_options, selected_video_names=args.video_names
    )

    calibrator = _Calibrator(
        dataset=dataset,
        results_dir=calibration_results_dir,
        image_shape=dataset.image_shape,
        num_samples=args.num_samples,
        theta_dims=args.theta_dims,
    )

    writer = _ImagesAndScoresWriter() if kwargs["save_images"] else _ScoresWriter()

    for video_name in dataset.video_names:
        results_file = calibrator(video_name=video_name, writer=writer)
        yield video_name, results_file


def main():
    import argparse

    parser = argparse.ArgumentParser()

    # inputs
    parser.add_argument("dataset_loader", type=str)
    parser.add_argument("dataset_path", type=str)
    add_resizing_arguments(parser)
    parser.add_argument(
        "--video_names",
        type=str,
        nargs="+",
        help="Only evaluate using a subset of videos. " "If not set, will include all videos in dataset_path.",
    )
    parser.add_argument(
        "--num_samples",
        type=int,
        help="How many frames to perform the calibration in order to evaluate the image metric",
    )
    parser.add_argument("--theta_dims", type=int)
    parser.add_argument("--work_dir", type=str, help="Root folder for all experiments")
    parser.add_argument(
        "--save_images", default=False, action="store_true", help="Also save the images used for the calibration"
    )
    parser.add_argument("--random_seed", type=int, help="Optional random seed for deterministic results")

    args = parser.parse_args()

    list(calibrate(**vars(args)))


if __name__ == "__main__":
    main()



"""
Generates synthetic data using multiprocessing
"""

import os
import pickle
import time
from multiprocessing import Process, Value
from typing import Type, Callable, Generator, Optional

import numpy as np

from wormpose.dataset import Dataset
from wormpose.machine_learning.generic_file_writer import GenericFileWriter
from wormpose.pose.centerline import flip_theta


class _TemplatesChunkData(object):
    def __init__(self, num_total_templates):

        self.frames = None
        self.measurements = None
        self.skeletons = None
        self.video_names = None

        self._num_total = num_total_templates
        self._cur_index = 0

    def _init_chunk_arr(self, arr):
        arr = np.asarray(arr)
        return np.empty((self._num_total,) + arr.shape[1:], dtype=arr.dtype)

    def add_data(self, features, frames, template_indexes, video_name: str):

        # sort indexes for faster frames reading
        template_indexes = np.sort(template_indexes)

        if self._cur_index == 0:
            # init all chunk arrays
            # frames images may not have all the same size, so preallocate a simple python list instead
            self.frames = [None] * self._num_total
            self.skeletons = self._init_chunk_arr(features.skeletons)
            self.measurements = self._init_chunk_arr(features.measurements)
            self.video_names = self._init_chunk_arr(video_name)

        for i in range(len(template_indexes)):
            template_index = template_indexes[i]

            self.frames[self._cur_index] = frames[template_index]
            self.skeletons[self._cur_index] = features.skeletons[template_index]
            self.measurements[self._cur_index] = features.measurements[template_index]
            self.video_names[self._cur_index] = video_name

            self._cur_index = self._cur_index + 1

    def write_to_file(self, filename: str):
        with open(filename, "wb") as f:
            pickle.dump(self, f)


def _write_to_file(
    out_filename: str,
    num_samples: int,
    template_filename: str,
    postures_generation_fn: Callable[[], Generator],
    progress_counter,
    writer: Type[GenericFileWriter],
    synthetic_dataset_args,
    random_seed: Optional[int],
):
    from wormpose.images.synthetic import SyntheticDataset

    seed = int.from_bytes(os.urandom(4), byteorder="little") if random_seed is None else random_seed
    np.random.seed(seed)

    with open(template_filename, "rb") as templates_f:
        templates = pickle.load(templates_f)
    os.remove(template_filename)

    worm_measurements = {}
    for video_name in np.unique(templates.video_names):
        indexes = np.where(templates.video_names == video_name)[0]
        measurements = templates.measurements[indexes]
        average_measurements = np.empty((measurements.shape[1:]), dtype=measurements.dtype)
        for name in measurements.dtype.names:
            average_measurements[0][name] = np.nanmean(measurements[name])
        worm_measurements[video_name] = average_measurements

    synthetic_dataset = SyntheticDataset(**synthetic_dataset_args)

    # preallocate all the template indexes to create the synthetic images (random)
    template_indexes = np.random.randint(len(templates.frames), size=num_samples)

    # preallocate all the choices for the position of the head (random)
    headtail_choice = np.random.choice([0, 1], size=num_samples)

    # preallocate image buffer to draw the synthetic image on
    image_data = np.empty(synthetic_dataset.output_image_shape, dtype=np.uint8)

    postures_gen = postures_generation_fn()
    with writer(out_filename) as synth_data_writer:

        for index, (cur_template_index, cur_headtail_choice) in enumerate(zip(template_indexes, headtail_choice)):
            theta = next(postures_gen)
            label_data = [theta, flip_theta(theta)]
            video_name = templates.video_names[cur_template_index]
            template_frame = templates.frames[cur_template_index]
            template_skeleton = templates.skeletons[cur_template_index]
            template_measurements = worm_measurements[video_name]
            synthetic_dataset.generate(
                theta=label_data[cur_headtail_choice],
                template_frame=template_frame,
                template_skeleton=template_skeleton,
                template_measurements=template_measurements,
                out_image=image_data,
            )
            synth_data_writer.write(locals())
            progress_counter.value = index + 1


class SyntheticDataGenerator(object):
    """
    This generator makes synthetic data by distributing the work to several processes:
    Each process will write one file containing the synthetic data and its label
    We use a "GenericFileWriter" to be able to write the files in different formats (TFrecord, pickle etc...)
    """

    _MAX_TEMPLATE_FRAMES = 1000

    def __init__(
        self,
        num_process: int,
        temp_dir: str,
        dataset: Dataset,
        postures_generation_fn: Callable[[], Generator],
        enable_random_augmentations: bool,
        writer: Type[GenericFileWriter],
        random_seed: Optional[int],
    ):
        """

        :param num_process: How many processes to distribute the work: there will be one file written per process
        :param temp_dir: Where to store temporary files
        :param dataset: A WormPose dataset
        :param postures_generation_fn: A function generating worm postures (angles)
        :param enable_random_augmentations: If true, the synthetic images will have random augmentations:
            translation, scale, blur
        :param writer: object responsible of writing the data to files
        :param random_seed: Seed for reproducibility
        """

        self.num_process = num_process
        self.temp_dir = temp_dir
        self.dataset = dataset
        self.postures_generation_fn = postures_generation_fn
        self.writer = writer
        self.random_seed = random_seed

        self.synthetic_dataset_args = dict(
            frame_preprocessing=dataset.frame_preprocessing,
            enable_random_augmentations=enable_random_augmentations,
            output_image_shape=dataset.image_shape,
        )

        self.video_names = dataset.video_names
        self._all_template_indexes = [dataset.features_dataset[v].labelled_indexes for v in self.video_names]

        if np.sum([len(x) for x in self._all_template_indexes]) == 0:
            raise RuntimeError("Can't create training data because couldn't find any labelled frame in the dataset.")

    def _split_data(self, num_samples_per_process):
        """
         Split a subset of template frames and associated features from different videos randomly to several processes
         """
        for process_index in range(self.num_process):

            num_samples = num_samples_per_process[process_index]
            if num_samples == 0:
                continue

            num_templates = min(num_samples, self._MAX_TEMPLATE_FRAMES)

            projected_num_templates_per_video = max(num_templates // len(self.video_names), 1)
            actual_num_templates_per_video = [
                min(projected_num_templates_per_video, len(x)) for x in self._all_template_indexes
            ]
            num_total_templates = np.sum(actual_num_templates_per_video)
            templates_chunk_data = _TemplatesChunkData(num_total_templates=num_total_templates)

            for video_name, cur_video_template_indexes, cur_video_num_templates in zip(
                self.video_names, self._all_template_indexes, actual_num_templates_per_video,
            ):
                # Select randomly some template indexes for each video
                template_indexes = np.random.choice(
                    cur_video_template_indexes, size=cur_video_num_templates, replace=False,
                )
                # Fill in the templates data chunk for this video
                with self.dataset.frames_dataset.open(video_name) as frames:
                    templates_chunk_data.add_data(
                        features=self.dataset.features_dataset[video_name],
                        frames=frames,
                        template_indexes=template_indexes,
                        video_name=video_name,
                    )

            template_filename = os.path.join(self.temp_dir, f"template_frames_{process_index:02d}.pkl")
            templates_chunk_data.write_to_file(template_filename)

            yield template_filename

    def generate(self, file_pattern: str, num_samples: int):
        """
        Start the generation of files containing synthetic data

        :param file_pattern: Filepath of the files to generate with an "index" variable
            ex: "path_to/file_{index}.tfrecord"
        :param num_samples: How many images (total) to generate, they will be split in several files
        """

        if num_samples <= 0:
            return 1  # nothing to do when num_samples is <= 0, progress 100%

        progress_counters = [Value("i", 0) for _ in range(self.num_process)]

        num_samples_per_process = [num_samples // self.num_process] * self.num_process
        num_samples_per_process[0] += num_samples % self.num_process

        workers = []
        for process_index, template_filename in enumerate(self._split_data(num_samples_per_process)):
            seed = self.random_seed + process_index if self.random_seed is not None else None
            out_filename = file_pattern.format(index=process_index)
            proc = Process(
                target=_write_to_file,
                args=(
                    out_filename,
                    num_samples_per_process[process_index],
                    template_filename,
                    self.postures_generation_fn,
                    progress_counters[process_index],
                    self.writer,
                    self.synthetic_dataset_args,
                    seed,
                ),
            )
            proc.start()
            workers.append(proc)

        while True:
            num_entries_written = np.sum([x.value for x in progress_counters])
            yield float(num_entries_written) / num_samples

            if num_entries_written >= num_samples:
                break
            time.sleep(0.5)

        for proc in workers:
            proc.join()



"""
Loaders for Tierpsy tracker data

The files should be organized this way :

+-- dataset_path
|   +-- video_name0
|   |   +-- video_name0.hdf5
|   |   +-- video_name0_features.hdf5  or video_name0_featuresN.hdf5
|   +-- video_name1
|   |   +-- video_name1.hdf5
|   |   +-- video_name1_features.hdf5 or video_name1_featuresN.hdf5
|   ...

WARNING not all configurations of Tierpsy tracker are supported.
For example: only one worm per file will be extracted (the one with the smallest index)
"""

from wormpose.dataset.image_processing.simple_frame_preprocessing import SimpleFramePreprocessing

FramePreprocessing = SimpleFramePreprocessing
from wormpose.dataset.loaders.tierpsy.frames_dataset import FramesDataset
from wormpose.dataset.loaders.tierpsy.features_dataset import FeaturesDataset
from wormpose.dataset.loaders.tierpsy.results_exporter import ResultsExporter



#!/usr/bin/env python

"""
Visualizer for the synthetic images
"""

import random
from typing import Optional, Generator
import numpy as np

from wormpose.dataset.loader import load_dataset
from wormpose.dataset.loaders.resizer import add_resizing_arguments, ResizeOptions
from wormpose.images.synthetic import SyntheticDataset
from wormpose.pose.postures_model import PosturesModel


class SyntheticSimpleVisualizer(object):
    """
    Utility class to visualize the synthetic images
    """

    def __init__(
        self,
        dataset_loader: str,
        dataset_path: str,
        postures_generator: Optional[Generator] = None,
        video_name: str = None,
        **kwargs
    ):
        resize_options = ResizeOptions(**kwargs)
        dataset = load_dataset(dataset_loader, dataset_path, resize_options=resize_options)

        if postures_generator is None:
            postures_generator = PosturesModel().generate()
        if video_name is None:
            video_name = dataset.video_names[0]

        features = dataset.features_dataset[video_name]
        self.skeletons = features.skeletons
        self.measurements = features.measurements

        self.output_image_shape = dataset.image_shape

        self.synthetic_dataset = SyntheticDataset(
            frame_preprocessing=dataset.frame_preprocessing,
            output_image_shape=self.output_image_shape,
            enable_random_augmentations=False,
        )
        skel_is_not_nan = ~np.any(np.isnan(self.skeletons), axis=(1, 2))
        self.labelled_indexes = np.where(skel_is_not_nan)[0]
        if len(self.labelled_indexes) == 0:
            raise ValueError("No template frames found in the dataset, can't generate synthetic images.")
        self.frames_dataset = dataset.frames_dataset
        self.video_name = video_name
        self.postures_generator = postures_generator

    def generate(self):
        out_image = np.empty(self.output_image_shape, dtype=np.uint8)

        with self.frames_dataset.open(self.video_name) as frames:
            while True:
                theta = next(self.postures_generator)
                random_label_index = np.random.choice(self.labelled_indexes)
                self.synthetic_dataset.generate(
                    theta=theta,
                    template_skeleton=self.skeletons[random_label_index],
                    template_frame=frames[random_label_index],
                    out_image=out_image,
                    template_measurements=self.measurements,
                )
                yield out_image, theta


def main():
    import argparse
    import cv2

    parser = argparse.ArgumentParser()
    parser.add_argument("dataset_loader", type=str)
    parser.add_argument("dataset_path", type=str)
    parser.add_argument("--video_name", type=str)
    parser.add_argument("--random_seed", type=int, help="Optional random seed for deterministic results")
    add_resizing_arguments(parser)

    args = parser.parse_args()

    random.seed(args.random_seed)
    np.random.seed(args.random_seed)

    synth_visualizer_gen = SyntheticSimpleVisualizer(**vars(args)).generate()

    while True:
        synth_image, _ = next(synth_visualizer_gen)

        cv2.imshow("synth_image", synth_image)
        cv2.waitKey()


if __name__ == "__main__":
    main()



"""
Dataset loader for the tutorial sample data
"""

from wormpose.dataset.image_processing.simple_frame_preprocessing import SimpleFramePreprocessing
from wormpose.dataset.loaders.hdf5features import HDF5Features
from wormpose.dataset.loaders.images_folder import ImagesFolder

FramesDataset = ImagesFolder
FeaturesDataset = HDF5Features
FramePreprocessing = SimpleFramePreprocessing



"""
 Image processing example implementation for segmenting the worm in an image
"""

from typing import Callable

import cv2
import numpy as np

# how much to crop the image to find the biggest blob
# (this is to avoid picking a biggest blob that is not the worm)
_CROP_PERCENT = 0.15


class ConstantThreshold:
    """
    Threshold function that always returns the same threshold
    """

    def __init__(self, threshold_value):
        self.threshold_value = threshold_value

    def __call__(self, frame: np.ndarray) -> int:
        return self.threshold_value


class OtsuThreshold(object):
    """
    Calculates automatic Otsu threshold on the blurred frame
    """

    def __init__(self, blur_kernel):
        """
        Creates an Otsu threshold operation with a preprocessing gaussian blur

        :param blur_kernel: Gaussian Kernel Size for the blur operation before the Otsu threshold method
            to split background and foreground. [height width]. height and width should be odd and can have different values.
        """
        self.blur_kernel = blur_kernel

    def __call__(self, frame: np.ndarray) -> int:
        blurred_frame = cv2.GaussianBlur(frame, self.blur_kernel, 0)
        blurred_frame[frame == 0] = 0
        background_threshold, _ = cv2.threshold(
            blurred_frame[blurred_frame > 0], 0, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU,
        )
        return background_threshold


def segment_foreground(
    frame: np.ndarray,
    foreground_close_struct_element,
    foreground_dilate_struct_element,
    threshold_fn: Callable[[np.ndarray], int],
):
    """
    Processes a frame to isolate the object of interest (worm) from the background

    :param frame: image to process
    :param foreground_close_struct_element: morphological element to close holes in the foreground mask
    :param foreground_dilate_struct_element: morphological element to expand the foreground mask
    :param threshold_fn: function that will return the threshold to separate forefround from background in a frame
    :return: segmentation mask with values of 1 for the worm object and 0 for the background,
        and average value of the background pixels
    """

    # find the threshold to separate foreground from background
    background_threshold = threshold_fn(frame)

    # use the threshold to deduce background and foreground masks, fill in holes
    foreground_mask = (frame > 0).astype(np.uint8) * (frame < background_threshold).astype(np.uint8)
    foreground_mask = cv2.morphologyEx(foreground_mask, cv2.MORPH_CLOSE, foreground_close_struct_element)
    background_mask = ((frame > 0).astype(np.uint8) - foreground_mask) > 0

    # calculate the average background color
    background_values = frame[background_mask]
    background_color = int(np.mean(background_values)) if len(background_values) > 0 else 0
    background_color = frame.dtype.type(background_color)

    # process the foreground mask to eliminate non worm objects
    # use connected components to find blobs, but focus on the center of the image to find the biggest
    # modify foreground_mask to only show the worm object
    nb_labels, labels, stats, _ = cv2.connectedComponentsWithStats(foreground_mask)
    labels_crop_size = int(_CROP_PERCENT * max(foreground_mask.shape))
    labels_cropped = labels[
        labels_crop_size : foreground_mask.shape[0] - labels_crop_size,
        labels_crop_size : foreground_mask.shape[1] - labels_crop_size,
    ]

    if nb_labels == 1:
        foreground_mask.fill(0)

    foreground_objects_sizes = [len(np.where(labels_cropped == l)[0]) for l in range(1, nb_labels)]
    if len(foreground_objects_sizes) > 0:
        biggest_blob_label = np.argmax(foreground_objects_sizes) + 1
        foreground_mask[labels != biggest_blob_label] = 0

    # add a little padding to the foreground mask
    foreground_mask = cv2.dilate(foreground_mask, foreground_dilate_struct_element)

    return foreground_mask, background_color



"""
This module contains function related to the worm centerline, for example to convert skeleton (x, y) coordinates
to angles, flip the head-tail orientation etc
"""
from typing import Tuple, Optional

import numpy as np
from scipy.interpolate import interp1d


def calculate_skeleton(
    theta: np.ndarray,
    worm_length: float,
    out: Optional[np.ndarray] = None,
    canvas_width_height: Optional[Tuple[int, int]] = None,
) -> np.ndarray:
    """
    Calculates (x,y) coordinates of the worm centerline from the tangent angle (theta) vector

    :param theta: centerline angles
    :param worm_length: desired worm length
    :param out: optional numpy array of shape (len(theta), 2) to store the resulting skeleton
    :param canvas_width_height: optional canvas to recenter the skeleton coordinates in the middle
    :return: output skeleton: numpy array of shape (len(theta), 2) representing the centerline angles
    """
    num_centerline_joints = len(theta)

    if out is None:
        out = np.empty((num_centerline_joints, 2))

    centerline_section_length = worm_length / (num_centerline_joints - 1)
    np.cumsum(
        [centerline_section_length * np.cos(theta), centerline_section_length * np.sin(theta),], axis=1, out=out.T,
    )

    # optional center the skeleton coordinates in the middle of an image canvas
    if canvas_width_height is not None:
        out += -np.min(out, axis=0) + (canvas_width_height - (np.max(out, axis=0) - np.min(out, axis=0))) / 2

    return out


def interpolate_skeleton(skeleton: np.ndarray, new_dims: int) -> np.ndarray:
    """
    Interpolates a worm skeleton to have a different number of points
    """
    new_pos_dim = []
    for dim in range(skeleton.shape[1]):
        y = skeleton[:, dim]
        x = np.arange(y.size)
        if np.any(np.isnan(y)):
            new_pos_dim.append([np.nan] * (new_dims + 1))
        else:
            # Interpolate the data using a cubic spline to "new_length" samples
            new_length = new_dims + 1
            new_x = np.linspace(x.min(), x.max(), new_length)
            new_y = interp1d(x, y, kind="cubic")(new_x)
            new_pos_dim.append(new_y)
    new_pos = np.vstack(new_pos_dim).T
    return new_pos


def skeletons_to_angles(skeletons: np.ndarray, theta_dims: int) -> np.ndarray:
    new_skeletons = []
    for frame in range(skeletons.shape[0]):
        skeleton = skeletons[frame]
        new_skeletons.append(interpolate_skeleton(skeleton, theta_dims))
    new_skeletons = np.array(new_skeletons, skeletons.dtype)

    skel_x = new_skeletons[:, :, 0]
    skel_y = new_skeletons[:, :, 1]
    d_x = np.diff(skel_x, axis=1)
    d_y = np.diff(skel_y, axis=1)
    # calculate tangent angles.  atan2 uses angles from -pi to pi
    angles = np.arctan2(d_y, d_x)
    return angles.astype(np.float32)


def skeleton_to_angle(skeleton: np.ndarray, theta_dims: int):
    new_skeleton = interpolate_skeleton(skeleton, theta_dims)
    skel_x = new_skeleton[:, 0]
    skel_y = new_skeleton[:, 1]
    d_x = np.diff(skel_x)
    d_y = np.diff(skel_y)
    # calculate tangent angles.  atan2 uses angles from -pi to pi
    angles = np.arctan2(d_y, d_x)
    return angles.astype(np.float32)


def get_joint_indexes(nb_skeleton_joints: int) -> Tuple[int, int, int]:
    head_joint = int(0.1 * nb_skeleton_joints)
    mid_body_joint = int(0.5 * nb_skeleton_joints)
    tail_body_joint = int(0.9 * nb_skeleton_joints)

    return head_joint, mid_body_joint, tail_body_joint


def flip_theta_series(theta_series: np.ndarray) -> np.ndarray:
    """
     head-tail flip for each value in a serie
    """
    return np.flip(theta_series, axis=1) + np.pi


def flip_theta(theta: np.ndarray) -> np.ndarray:
    """
     head-tail flip for one value
    """
    return np.flip(theta, axis=0) + np.pi



import cv2
import numpy as np

from wormpose.images.worm_drawing import make_draw_worm_body
from wormpose.pose.centerline import calculate_skeleton


class TargetSkeletonCalculator(object):
    """
    Calculates the target skeleton
    Makes sure the target skeleton has the same dimensions as the template skeleton by interpolation if necessary
    """

    def __init__(self, nb_skeletons_joints, theta_dims, output_image_shape):
        self.output_image_shape = output_image_shape
        self.need_interpolation = nb_skeletons_joints != theta_dims
        if self.need_interpolation:
            self.interp_x = np.linspace(0, theta_dims - 1, nb_skeletons_joints)
            self.interp_xp = np.arange(0, theta_dims)
        self.non_interpolated_skeleton = np.empty((theta_dims, 2), dtype=float)

    def calculate(self, theta, worm_length, out):

        calculate_skeleton(
            theta, worm_length, canvas_width_height=self.output_image_shape, out=self.non_interpolated_skeleton,
        )

        if self.need_interpolation:
            self._interpolate(self.non_interpolated_skeleton[:, 0], dst=out[:, 0])
            self._interpolate(self.non_interpolated_skeleton[:, 1], dst=out[:, 1])
        else:
            np.copyto(dst=out, src=self.non_interpolated_skeleton)

    def _interpolate(self, non_interpolated, dst):
        np.copyto(dst=dst, src=np.interp(self.interp_x, self.interp_xp, non_interpolated))


class WormOutlineMask(object):
    """
    Filters an image by a mask representing the global outline of the worm,
    to smooth out the rough edges of assembling the body segment image patches
    """

    def __init__(self, output_image_shape):
        self.draw_worm_body = make_draw_worm_body()
        self.target_shape_outline_image = np.empty(output_image_shape, dtype=np.uint8)

    def apply(self, worm_thickness, output_image, target_skel):
        # reset mask
        self.target_shape_outline_image.fill(0)

        # draw the target worm outline on the mask
        self.draw_worm_body(worm_thickness, self.target_shape_outline_image, skeleton=target_skel)

        # filter the synthetic image with the mask of the worm outline
        output_image[self.target_shape_outline_image == 0] = 0


class PatchDrawing(object):
    """
    Transform image patches of body segments between a template image to a target image
    """

    def __init__(self, output_image_shape):
        self.source_image_mask = None
        self.output_image_shape = output_image_shape

        # preallocate arrays for image processing operations
        self.target_image_mask = np.empty(self.output_image_shape, dtype=np.uint8)
        self.cur_alpha = np.empty(self.output_image_shape, dtype=np.float32)
        self.all_alpha = np.empty(self.output_image_shape, dtype=np.float32)
        self.target_image = np.empty(self.output_image_shape, dtype=np.uint8)

    def draw_patch(self, template_coords, target_coords, recentered_frame, out_image):
        # grow preallocated image as necessary
        if (
            self.source_image_mask is None
            or self.source_image_mask.shape[0] < recentered_frame.shape[0]
            or self.source_image_mask.shape[1] < recentered_frame.shape[1]
        ):
            self.source_image_mask = np.empty_like(recentered_frame)
        source_image_mask = self.source_image_mask[: recentered_frame.shape[0], : recentered_frame.shape[1]]

        transform = cv2.getAffineTransform(target_coords[:3], template_coords[:3])
        source_image_mask.fill(0)
        cv2.fillConvexPoly(source_image_mask, target_coords.astype(np.int), 255)
        masked_frame = cv2.bitwise_and(recentered_frame, source_image_mask)
        cv2.warpAffine(
            masked_frame,
            transform,
            dsize=(self.output_image_shape[1], self.output_image_shape[0]),
            borderMode=cv2.BORDER_CONSTANT,
            flags=cv2.INTER_NEAREST,
            dst=self.target_image,
        )
        self.target_image_mask.fill(0)
        cv2.fillConvexPoly(self.target_image_mask, template_coords.astype(np.int), 255)
        cv2.bitwise_and(self.target_image, self.target_image_mask, dst=self.target_image)

        np.greater(self.target_image, 0, out=self.cur_alpha)
        np.greater(out_image, 0, out=self.all_alpha)
        self.all_alpha += self.cur_alpha
        self.all_alpha[self.all_alpha > 1] = 0.5
        out_image += self.target_image
        out_image *= self.all_alpha



#!/usr/bin/env python

"""
Trains the neural network on the training data, supports resuming training
"""

import glob
import logging
import multiprocessing as mp
import os
import random
from argparse import Namespace

import numpy as np
import tensorflow as tf

from wormpose.commands import _log_parameters
from wormpose.config import default_paths
from wormpose.config.default_paths import SYNTH_TRAIN_DATASET_NAMES, REAL_EVAL_DATASET_NAMES, CONFIG_FILENAME
from wormpose.config.experiment_config import load_config, add_config_argument
from wormpose.dataset.loader import get_dataset_name
from wormpose.machine_learning import model
from wormpose.machine_learning.best_models_saver import BestModels
from wormpose.machine_learning.loss import symmetric_angle_difference
from wormpose.machine_learning.tfrecord_file import get_tfrecord_dataset

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
tf.get_logger().setLevel(logging.INFO)


def _find_tfrecord_files(experiment_dir: str):
    training_data_dir = os.path.join(experiment_dir, default_paths.TRAINING_DATA_DIR)
    train_tfrecord_filenames = glob.glob(os.path.join(training_data_dir, SYNTH_TRAIN_DATASET_NAMES.format(index="*")))
    eval_tfrecord_filenames = glob.glob(os.path.join(training_data_dir, REAL_EVAL_DATASET_NAMES.format(index="*")))
    if len(train_tfrecord_filenames) == 0 or len(eval_tfrecord_filenames) == 0:
        raise FileNotFoundError("Training/Eval dataset not found.")
    return train_tfrecord_filenames, eval_tfrecord_filenames


def _parse_arguments(dataset_path: str, kwargs: dict):
    if kwargs.get("work_dir") is None:
        kwargs["work_dir"] = default_paths.WORK_DIR
    if kwargs.get("batch_size") is None:
        kwargs["batch_size"] = 128
    if kwargs.get("epochs") is None:
        kwargs["epochs"] = 100
    if kwargs.get("network_model") is None:
        kwargs["network_model"] = model.build_model
    if kwargs.get("optimizer") is None:
        kwargs["optimizer"] = "adam"
    if kwargs.get("loss") is None:
        kwargs["loss"] = symmetric_angle_difference
    if kwargs.get("random_seed") is None:
        kwargs["random_seed"] = None

    dataset_name = get_dataset_name(dataset_path)
    kwargs["experiment_dir"] = os.path.join(kwargs["work_dir"], dataset_name)

    if kwargs.get("config") is None:
        kwargs["config"] = os.path.join(kwargs["experiment_dir"], CONFIG_FILENAME)

    _log_parameters(logger.info, {"dataset_path": dataset_path})
    _log_parameters(logger.info, kwargs)

    return Namespace(**kwargs)


def train(dataset_path: str, **kwargs):
    """
    Train a neural network with the TFrecord files generated with the script generate_training_data
    Save the best model performing on evaluation data

    :param dataset_path: Root path of the dataset containing videos of worm
    """
    args = _parse_arguments(dataset_path, kwargs)

    mp.set_start_method("spawn", force=True)

    if args.random_seed is not None:
        os.environ["TF_DETERMINISTIC_OPS"] = "1"
        random.seed(args.random_seed)
        np.random.seed(args.random_seed)
        tf.random.set_seed(args.random_seed)

    models_dir = os.path.join(args.experiment_dir, default_paths.MODELS_DIRS)
    if not os.path.exists(models_dir):
        os.mkdir(models_dir)

    train_tfrecord_filenames, eval_tfrecord_filenames = _find_tfrecord_files(args.experiment_dir)

    config = load_config(args.config)
    if config.num_eval_samples < args.batch_size or config.num_train_samples < args.batch_size:
        raise ValueError("The number of samples in the train and eval datasets must be higher than the batch size.")

    train_dataset = get_tfrecord_dataset(
        filenames=train_tfrecord_filenames,
        image_shape=config.image_shape,
        batch_size=args.batch_size,
        theta_dims=config.theta_dimensions,
        is_train=True,
    )
    validation_dataset = get_tfrecord_dataset(
        filenames=eval_tfrecord_filenames,
        image_shape=config.image_shape,
        batch_size=args.batch_size,
        theta_dims=config.theta_dimensions,
        is_train=False,
    )

    tensorboard_callback = tf.keras.callbacks.TensorBoard(
        log_dir=os.path.join(args.experiment_dir, "tensorboard_log"), histogram_freq=1
    )
    best_models_callback = BestModels(models_dir)
    checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
        best_models_callback.models_name_pattern,
        save_best_only=False,
        save_weights_only=False,
        monitor="val_loss",
        mode="min",
    )

    keras_model = args.network_model(input_shape=config.image_shape, out_dim=config.theta_dimensions)
    last_model_path = best_models_callback.last_model_path
    if os.path.isfile(last_model_path):
        keras_model = tf.keras.models.load_model(last_model_path, compile=False)

    keras_model.compile(optimizer=args.optimizer, loss=args.loss)
    keras_model.fit(
        train_dataset,
        epochs=args.epochs,
        steps_per_epoch=config.num_train_samples // args.batch_size,
        shuffle=False,
        initial_epoch=best_models_callback.epoch,
        validation_data=validation_dataset,
        validation_steps=config.num_eval_samples // args.batch_size,
        callbacks=[tensorboard_callback, checkpoint_callback, best_models_callback],
    )


def main():
    import argparse

    parser = argparse.ArgumentParser()

    parser.add_argument("dataset_path", type=str)

    add_config_argument(parser)
    parser.add_argument("--batch_size", type=int, help="Batch size for training")
    parser.add_argument("--epochs", type=int, help="How many epochs to train the network")
    parser.add_argument("--work_dir", type=str, help="Root folder for all experiments")
    parser.add_argument("--optimizer", type=str, help="Which optimizer for training, 'adam' by default.")
    parser.add_argument("--random_seed", type=int, help="Optional random seed for deterministic results")

    args = parser.parse_args()

    train(**vars(args))


if __name__ == "__main__":
    main()



"""
Assesses the accuracy of a centerline theta compared to a real image, by calculating the image similarity metric
"""

import numpy as np

from wormpose import BaseFramePreprocessing
from wormpose.images.scoring import image_scoring
from wormpose.images.scoring.image_scoring import fit_bounding_box_to_worm
from wormpose.images.real_dataset import RealDataset
from wormpose.images.synthetic import SyntheticDataset


class CenterlineAccuracyCheck(object):
    """
    Class that performs comparison (image similarity) to assess how a centerline accurately represents a real image.

    It first preprocesses the original real image with a BaseFramePreprocessing class
    (crop and set the background pixels to a uniform color).
    Then, it creates a synthetic image representing the centerline theta, using a provided template image.
    Typically, the template image was chosen to be the closest labelled image in time to the real image.
    The synthetic image is cropped to fit the worm, in order to apply a template matching function between
    the real image (full size) and the synthetic image (smaller)
    The result is an image similarity value and the synthetic image skeleton coordinates.
    """

    def __init__(self, frame_preprocessing: BaseFramePreprocessing, image_shape):

        self.real_dataset = RealDataset(frame_preprocessing=frame_preprocessing, output_image_shape=image_shape)

        self.synthetic_dataset = SyntheticDataset(
            frame_preprocessing=frame_preprocessing, output_image_shape=image_shape, enable_random_augmentations=False,
        )
        self.last_synth_image = np.empty(image_shape, np.uint8)
        self.last_real_image = None

    def __call__(
        self, theta, template_skeleton, template_frame, template_measurements, real_frame_orig,
    ):
        if np.any(np.isnan(theta)):
            score = np.nan
            synth_skel = np.full_like(template_skeleton, np.nan)
            return score, synth_skel

        self.last_real_image, skel_offset = self.real_dataset.process_frame(real_frame_orig)
        cur_bg_color, synth_skel = self.synthetic_dataset.generate(
            theta,
            template_frame=template_frame,
            template_skeleton=template_skeleton,
            out_image=self.last_synth_image,
            template_measurements=template_measurements,
        )

        # Crop the synthetic image to the object of interest before doing the image comparison,
        # we don't need the full image with all the background, still keep a little padding around the worm.
        left, right, bottom, top = fit_bounding_box_to_worm(self.last_synth_image, cur_bg_color)
        np.subtract(synth_skel, (bottom, left), out=synth_skel)
        cropped_synth_image = self.last_synth_image[left:right, bottom:top]

        # Perform the image comparison between the real image and the reconstructed synthetic image cropped.
        # This gives a heat map, we are interested in the maximum value of the heatmap and its location.
        # This maximum score gives an estimation of the confidence of the prediction.
        score, score_loc = image_scoring.calculate_similarity(
            source_image=self.last_real_image, template_image=cropped_synth_image
        )

        # Using the heatmap maximum coordinates, we can transform the coordinates of the reconstructed skeleton
        # to the original image coordinates system.
        synth_skel += score_loc
        synth_skel += np.array([skel_offset[0], skel_offset[1]])

        return score, synth_skel



"""
This module deals with loading features from a dataset.

It will calculate extra features such as the worm length.
"""

from typing import Dict, Tuple

import numpy as np


class Features(object):
    def __init__(self, raw_features: dict):

        _validate_features(raw_features)

        self.skeletons = raw_features["skeletons"]
        self.frame_rate = raw_features["frame_rate"]
        self.timestamp = raw_features["timestamp"] if "timestamp" in raw_features else None
        self.ventral_side = raw_features["ventral_side"] if "ventral_side" in raw_features else None

        # set defaults for optional values
        if self.ventral_side is None:
            self.ventral_side = "unknown"
        if self.timestamp is None:  # default behavior: all frames are equidistant in time
            self.timestamp = np.arange(0, len(self.skeletons))

        self._set_measurements(raw_features)
        self._set_labelled_indexes()

    def _set_labelled_indexes(self):
        # Save indexes where skeleton is valid (not nan)
        skel_is_not_nan = ~np.any(np.isnan(self.skeletons), axis=(1, 2))
        self.labelled_indexes = np.where(skel_is_not_nan)[0]

    def _set_measurements(self, raw_features: dict):
        worm_length = _calculate_worm_length(self.skeletons)
        self.measurements = np.stack(
            [worm_length, raw_features["head_width"], raw_features["midbody_width"], raw_features["tail_width"],],
            axis=1,
        )
        self.measurements.dtype = {
            "names": ("worm_length", "head_width", "midbody_width", "tail_width"),
            "formats": (float, float, float, float),
        }


FeaturesDict = Dict[str, Features]


def _calculate_worm_length(skeletons):
    worm_length = np.full(len(skeletons), np.nan, dtype=float)
    where_skel_nan = np.where(~np.any(np.isnan(skeletons), axis=(1, 2)))
    for i in where_skel_nan[0]:
        skel = skeletons[i]
        worm_length[i] = np.sum(np.sqrt(np.sum((skel[:-1] - skel[1:]) ** 2, axis=1)))
    return worm_length


def _validate_features(raw_features: dict):
    if not (
        len(raw_features["skeletons"])
        == len(raw_features["head_width"])
        == len(raw_features["midbody_width"])
        == len(raw_features["tail_width"])
    ):
        raise ValueError("inconsistent features")
    if not (
        raw_features["head_width"].dtype == raw_features["midbody_width"].dtype == raw_features["tail_width"].dtype
    ) or not np.issubdtype(raw_features["head_width"].dtype, np.floating):
        raise TypeError("Body measurements type should be identical and floating point")
    if not np.issubdtype(raw_features["skeletons"].dtype, np.floating):
        raise TypeError("Skeleton type should be floating point")
    if len(raw_features["skeletons"].shape) != 3 or raw_features["skeletons"].shape[2] != 2:
        raise ValueError("Wrong skeleton shape")
    if raw_features["skeletons"].shape[1] < 20:
        raise UserWarning(
            "Low number of skeleton joints (< 20), consider interpolating to improve quality of synthetic images"
        )
    if raw_features.get("timestamp") is not None and len(raw_features["timestamp"]) != len(raw_features["skeletons"]):
        raise ValueError("Inconsistent timestamp")


def calculate_max_average_worm_length(features: FeaturesDict) -> float:
    """
    Calculates the average worm length from each video, and returns the maximum.
    
    :param features: A dictionary of Features
    :return: Biggest average worm length from all videos
    """
    return float(np.nanmax([np.nanmean(x.measurements["worm_length"]) for x in features.values()]))


MINIMUM_IMAGE_SIZE = 32


def calculate_crop_window_size(features: FeaturesDict) -> Tuple[int, int]:
    """
    Returns an image shape that is just big enough to view the worm object,
    the image will be cropped (or expanded) to that size as an input to the neural network
    Can be overriden in child classes for another behavior (for example a fixed chosen size)
    :param features: A dictionary of Features
    :return: A tuple of two integer values (height, width)
    """
    import math

    # calculate the image shape as maximum of all the average worm lengths for each video
    # then we know the longest worm will fit the image shape,
    # the smaller ones will just have more background space
    max_average_worm_length = calculate_max_average_worm_length(features)
    if np.isnan(max_average_worm_length):
        raise ValueError(
            "Can't calculate the crop window size: "
            "couldn't get the max average worm length in the dataset."
            " Please check the labeled features in this dataset."
        )

    crop_size = max(MINIMUM_IMAGE_SIZE, int(max_average_worm_length))

    # round to even number
    crop_size = math.ceil(float(crop_size) / 2) * 2
    return crop_size, crop_size



"""
Applies safely the frame preprocessing function to a frame,
set the background pixels to a uniform value, deduces the region of interest
"""
from typing import Tuple

import numpy as np

from wormpose.dataset.base_dataset import BaseFramePreprocessing


def run(frame_preprocessing: BaseFramePreprocessing, frame: np.ndarray) -> Tuple[np.ndarray, int, Tuple[slice, slice]]:
    """
    Safely preprocesses an image, set the background pixels to a uniform color, calculates worm region of interest

    :param frame_preprocessing: Frame preprocessing logic
    :param frame: Image to preprocess
    :return: Processed image, value of the background color, region of interest coordinates
    """

    # copy to avoid modifying the source image
    frame_copy = np.copy(frame)

    # call the frame preprocessing function to get the segmented image
    segmentation_mask, background_color = frame_preprocessing.process(frame_copy)

    # enforces background color type
    background_color = int(background_color)

    # erase background, set everything not the foreground to a uniform color
    frame_copy[segmentation_mask == 0] = background_color

    # get region of interest (full image if no worm is found)
    where_worm = np.where(segmentation_mask != 0)
    if len(where_worm[0]) == 0 or len(where_worm[1]) == 0:
        worm_roi = np.s_[0 : frame_copy.shape[0], 0 : frame_copy.shape[1]]
    else:
        worm_roi = np.s_[
            np.min(where_worm[0]) : np.max(where_worm[0]), np.min(where_worm[1]) : np.max(where_worm[1]),
        ]

    return frame_copy, background_color, worm_roi



"""
Functions to read and write TFrecord files containing wormpose labeled data:
 one image paired with centerline angles (two options for head-tail flip)
"""

from functools import partial
from typing import Tuple, List

import tensorflow as tf

from wormpose.machine_learning.generic_file_writer import GenericFileWriter


def get_tfrecord_dataset(
    filenames, image_shape: Tuple[int, int], batch_size: int, theta_dims: int, is_train: bool,
):
    dataset = tf.data.TFRecordDataset(filenames, compression_type="GZIP")

    if is_train:
        dataset = dataset.shuffle(buffer_size=10000)
        dataset = dataset.repeat()

    dataset = dataset.map(
        partial(parse_example_normalize_image, theta_dims=theta_dims, image_shape=image_shape,),
        num_parallel_calls=tf.data.experimental.AUTOTUNE,
    )
    dataset = dataset.batch(batch_size)
    dataset = dataset.prefetch(buffer_size=tf.data.experimental.AUTOTUNE)
    return dataset


def parse_example(example, theta_dims: int):
    features = {
        "data": tf.io.FixedLenFeature([], tf.string),
        "label0": tf.io.FixedLenFeature([theta_dims], tf.float32),
        "label1": tf.io.FixedLenFeature([theta_dims], tf.float32),
    }
    parsed_features = tf.io.parse_single_example(example, features)
    return parsed_features


def parse_example_normalize_image(example, theta_dims: int, image_shape: Tuple[int, int]):
    parsed_features = parse_example(example, theta_dims)

    img = tf.io.decode_raw(parsed_features["data"], tf.uint8)
    img = tf.reshape(img, (image_shape[0], image_shape[1]))
    img = img[:, :, tf.newaxis]
    img = tf.cast(img, tf.float32)
    img /= 255.0

    labels = [parsed_features["label0"], parsed_features["label1"]]

    return img, labels


class Writer(object):
    def __init__(self, filename: str):
        self.filename = filename
        self.record_writer = None

    def __enter__(self):
        options = tf.io.TFRecordOptions(compression_type="GZIP")
        self.record_writer = tf.io.TFRecordWriter(self.filename, options=options)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.record_writer.close()

    def write(self, image_data, label_0, label_1):
        feature = {
            "data": tf.train.Feature(bytes_list=tf.train.BytesList(value=[image_data.tostring()])),
            "label0": tf.train.Feature(float_list=tf.train.FloatList(value=label_0.tolist())),
            "label1": tf.train.Feature(float_list=tf.train.FloatList(value=label_1.tolist())),
        }
        example_proto = tf.train.Example(features=tf.train.Features(feature=feature))
        self.record_writer.write(example_proto.SerializeToString())


def parse_square_image(example, theta_dims: int):
    parsed_features = parse_example(example, theta_dims)
    img = tf.io.decode_raw(parsed_features["data"], tf.uint8)
    img_size = tf.math.sqrt(tf.cast(tf.shape(img)[0], tf.float32))
    img = tf.reshape(img, (img_size, img_size))
    return img, [parsed_features["label0"], parsed_features["label1"]]


def read(filename: str, theta_dims: int):
    """
    Read a tfrecord file where the images in the files have the same width and height
    """
    raw_dataset = tf.data.TFRecordDataset(filename, compression_type="GZIP")
    parsed_dataset = raw_dataset.map(partial(parse_square_image, theta_dims=theta_dims))

    for parsed_record in parsed_dataset:
        yield parsed_record


def write_training_data_to_tfrecord(f, image_data, label_data, **kwargs):
    f.write(image_data, label_data[0], label_data[1])


class TfrecordLabeledDataWriter(GenericFileWriter):
    def __init__(self, filename: str):
        open_file = partial(Writer, filename)
        write_file = lambda f, data: write_training_data_to_tfrecord(f, **data)

        super().__init__(open_file=open_file, write_file=write_file)



#!/usr/bin/env python

"""
Exporting WormPose results to another format, for example the WCON format for a Tierpsy dataset
"""

import glob
import logging
import os
from argparse import Namespace

import h5py

from wormpose.commands import _log_parameters
from wormpose.config import default_paths
from wormpose.config.default_paths import RESULTS_FILENAME, CONFIG_FILENAME
from wormpose.config.experiment_config import load_config, add_config_argument
from wormpose.dataset import get_dataset_name, load_dataset

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def _parse_arguments(dataset_path: str, kwargs: dict):
    if kwargs.get("work_dir") is None:
        kwargs["work_dir"] = default_paths.WORK_DIR
    if kwargs.get("video_names") is None:
        kwargs["video_names"] = None

    dataset_name = get_dataset_name(dataset_path)
    kwargs["experiment_dir"] = os.path.join(kwargs["work_dir"], dataset_name)

    if kwargs.get("config") is None:
        kwargs["config"] = os.path.join(kwargs["experiment_dir"], CONFIG_FILENAME)

    _log_parameters(logger.info, {"dataset_path": dataset_path})
    _log_parameters(logger.info, kwargs)

    return Namespace(**kwargs)


def export(dataset_path: str, **kwargs):
    """
    Export WormPose results into another format, depending on the dataset

    :param dataset_path: Root path of the dataset containing videos of worm
    """
    args = _parse_arguments(dataset_path, kwargs)

    results_root_dir = os.path.join(args.experiment_dir, default_paths.RESULTS_DIR)
    results_files = glob.glob(os.path.join(results_root_dir, "*", RESULTS_FILENAME))

    if len(results_files) == 0:
        raise FileNotFoundError(f"No results files found in path: '{results_root_dir}'")

    config = load_config(args.config)

    dataset = load_dataset(config.dataset_loader, dataset_path, selected_video_names=args.video_names)

    for results_file in results_files:
        try:
            with h5py.File(results_file, "r") as f:
                results_skeletons = f["resolved"]["skeletons"][:]
        except Exception:
            logger.error(f"Couldn't read results in file {results_file}")
            continue

        video_name = os.path.basename(os.path.dirname(results_file))
        logger.info(f'Exporting: "{video_name}"')

        dataset.results_exporter.export(
            dataset=dataset,
            video_name=video_name,
            results_skeletons=results_skeletons,
            out_dir=os.path.dirname(results_file),
        )


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("dataset_path", type=str)
    parser.add_argument("--work_dir", type=str, help="Root folder for all experiments")
    parser.add_argument(
        "--video_names",
        type=str,
        nargs="+",
        help="Only analyze a subset of videos. If not set, will analyze all videos in dataset_path.",
    )
    add_config_argument(parser)
    args = parser.parse_args()

    export(**vars(args))


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""
Visualizer for the real processed images
"""

from wormpose.dataset.loader import load_dataset
from wormpose.dataset.loaders.resizer import add_resizing_arguments, ResizeOptions
from wormpose.images.real_dataset import RealDataset


class RealSimpleVisualizer(object):
    """
    Utility class to visualize the real images processed
    """

    def __init__(self, dataset_loader: str, dataset_path: str, video_name=None, **kwargs):
        resize_options = ResizeOptions(**kwargs)
        dataset = load_dataset(dataset_loader, dataset_path, resize_options=resize_options)

        self.video_name = video_name if video_name is not None else dataset.video_names[0]
        self.real_dataset = RealDataset(
            frame_preprocessing=dataset.frame_preprocessing, output_image_shape=dataset.image_shape,
        )
        self.frames_dataset = dataset.frames_dataset

    def generate(self):
        with self.frames_dataset.open(self.video_name) as frames:
            for frame in frames:
                processed_frame, _ = self.real_dataset.process_frame(frame)
                yield frame, processed_frame


def main():
    import argparse
    import cv2

    parser = argparse.ArgumentParser()
    parser.add_argument("dataset_loader", type=str)
    parser.add_argument("dataset_path", type=str)
    parser.add_argument(
        "--video_name", type=str, help="Optional video name. If not set, only visualize one video.",
    )
    add_resizing_arguments(parser)
    args = parser.parse_args()

    real_visualizer = RealSimpleVisualizer(**vars(args))

    for orig, processed in real_visualizer.generate():
        cv2.imshow("orig", orig)
        cv2.imshow("processed", processed)
        cv2.waitKey()


if __name__ == "__main__":
    main()



"""
Implements a Keras callback to save the top N best models on evaluation data,
"""

import heapq
import json
import logging
import os

import tensorflow as tf

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class _BestModelsHeap(object):
    """
    Keeps track of the top N named objects with the smallest value
    uses a max-heap (heapq and inverting the sign of the value)
    """

    def __init__(self, values, names, maxlen):
        self.maxlen = maxlen
        self._queue = []

        for val, name in zip(values, names):
            heapq.heappush(self._queue, (-val, name))

    def append(self, value, name):

        if len(self._queue) < self.maxlen:
            heapq.heappush(self._queue, (-value, name))
        else:
            heapq.heappushpop(self._queue, (-value, name))

    def get_sorted(self):
        sorted_queue = list(reversed(sorted(self._queue)))
        return [-x[0] for x in sorted_queue], [x[1] for x in sorted_queue]


_NAME_PATTERN = "model.{epoch:02d}-{val_loss:.2f}.hdf5"


class BestModels(tf.keras.callbacks.Callback):
    def __init__(self, models_dir, models_to_keep=5):

        self.models_name_pattern = os.path.join(models_dir, _NAME_PATTERN)
        self.models_dir = models_dir
        self.backup_filepath = os.path.join(self.models_dir, "models_info.json")

        try:
            with open(self.backup_filepath, "r") as f:
                backup = json.load(f)
                val_loss = backup["best_val_loss"]
                names = backup["best_models"]
                epoch = backup["epoch"]
                last = backup["last_model"]
        except Exception:
            val_loss = []
            names = []
            epoch = 0
            last = ""

        self.epoch = epoch
        self.last_model_path = os.path.join(models_dir, last)
        self.best_model_path = os.path.join(models_dir, names[0]) if len(names) > 0 else self.last_model_path
        self.models_heap = _BestModelsHeap(values=val_loss, names=names, maxlen=models_to_keep)

        super().__init__()

    def on_epoch_end(self, epoch, logs=None):
        """
        At the end of each epoch, update the best models and save to json
        """

        val_loss = logs["val_loss"]

        last_model_path = os.path.basename(self.models_name_pattern.format(epoch=epoch + 1, val_loss=val_loss))
        self.models_heap.append(val_loss, last_model_path)

        best_val_loss, best_model_names = self.models_heap.get_sorted()

        for f in os.listdir(self.models_dir):
            if f not in best_model_names and f != os.path.basename(self.backup_filepath) and f != last_model_path:
                os.remove(os.path.join(self.models_dir, f))

        with open(self.backup_filepath, "w") as f:
            json.dump(
                {
                    "epoch": epoch + 1,
                    "best_val_loss": best_val_loss,
                    "best_models": best_model_names,
                    "last_model": last_model_path,
                },
                f,
                indent=4,
            )



#!/usr/bin/env python

"""
Visualize the WormPose results as images showing the centerline on top of the original image
"""

import logging
import os
import shutil
import tempfile
import time
from argparse import Namespace

import cv2
import h5py
import numpy as np

from wormpose.commands import _log_parameters
from wormpose.config import default_paths
from wormpose.config.default_paths import RESULTS_FILENAME, CONFIG_FILENAME
from wormpose.config.experiment_config import load_config, add_config_argument
from wormpose.dataset.loader import get_dataset_name, Dataset
from wormpose.dataset.loader import load_dataset
from wormpose.images.worm_drawing import draw_skeleton

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class _Visualizer(object):
    def __init__(
        self, dataset: Dataset, results_dir: str, temp_dir: str, draw_original: bool,
    ):
        self.dataset = dataset
        self.results_dir = results_dir
        self.temp_dir = temp_dir
        self.draw_original = draw_original

    def export_to_images(self, video_name: str, results_scores, results_skel):

        results_dir = os.path.join(self.results_dir, video_name)

        images_temp_dir = tempfile.mkdtemp(dir=self.temp_dir)
        features = self.dataset.features_dataset[video_name]
        timestamp = features.timestamp

        image_filename_format = "time_{{:0{}d}}_frame_{{:0{}d}}_score_{{:.2f}}.png".format(
            len(str(timestamp[-1])), len(str(len(results_scores)))
        )
        start = time.time()

        original_color = (255, 255, 255)
        predicted_color = (0, 255, 0)

        with self.dataset.frames_dataset.open(video_name) as frames:

            for cur_time, (score, skel) in enumerate(zip(results_scores, results_skel)):

                frame_index = np.where(timestamp == cur_time)[0]
                if len(frame_index) == 0:
                    continue
                cur_frame_index = frame_index[0]
                cur_raw_frame = cv2.cvtColor(frames[cur_frame_index], cv2.COLOR_GRAY2BGR)

                if self.draw_original:
                    # draw original skeleton in white
                    draw_skeleton(
                        cur_raw_frame, features.skeletons[cur_frame_index], original_color, original_color,
                    )
                # draw skeleton from network predictions in green
                draw_skeleton(cur_raw_frame, skel, predicted_color, predicted_color)

                cv2.imwrite(
                    os.path.join(images_temp_dir, image_filename_format.format(cur_time, cur_frame_index, score),),
                    cur_raw_frame,
                )

        shutil.make_archive(os.path.join(results_dir, "images_results"), "zip", images_temp_dir)
        shutil.rmtree(images_temp_dir)

        end = time.time()
        logger.info(f"Exported result images for {os.path.basename(results_dir)} " f"in {end - start:.1f}s")


def _parse_arguments(dataset_path: str, kwargs: dict):

    if kwargs.get("temp_dir") is None:
        kwargs["temp_dir"] = tempfile.gettempdir()
    if kwargs.get("draw_original") is None:
        kwargs["draw_original"] = True
    if kwargs.get("work_dir") is None:
        kwargs["work_dir"] = default_paths.WORK_DIR
    if kwargs.get("video_names") is None:
        kwargs["video_names"] = None
    if kwargs.get("results_file") is None:
        kwargs["results_file"] = RESULTS_FILENAME
    if kwargs.get("group_name") is None:
        kwargs["group_name"] = "resolved"
    kwargs["temp_dir"] = tempfile.mkdtemp(dir=kwargs["temp_dir"])

    dataset_name = get_dataset_name(dataset_path)
    kwargs["experiment_dir"] = os.path.join(kwargs["work_dir"], dataset_name)

    if kwargs.get("config") is None:
        kwargs["config"] = os.path.join(kwargs["experiment_dir"], CONFIG_FILENAME)

    _log_parameters(logger.info, {"dataset_path": dataset_path})
    _log_parameters(logger.info, kwargs)

    return Namespace(**kwargs)


def visualize(dataset_path: str, **kwargs):
    """
    Export prediction results as videos with a centerline overlay on top of the original images

    :param dataset_path: Root path of the dataset containing videos of worm
    """
    args = _parse_arguments(dataset_path, kwargs)

    results_dir = os.path.join(args.experiment_dir, default_paths.RESULTS_DIR)
    config = load_config(args.config)

    dataset = load_dataset(config.dataset_loader, dataset_path, selected_video_names=args.video_names)

    visualizer = _Visualizer(
        dataset=dataset, draw_original=args.draw_original, temp_dir=args.temp_dir, results_dir=results_dir,
    )

    for video_name in dataset.video_names:

        results_file = os.path.join(results_dir, video_name, args.results_file)
        if not os.path.exists(results_file):
            logger.error(f"No results file to analyze, file not found: '{results_file}'")
            continue

        with h5py.File(results_file, "r") as f:
            if args.group_name not in f:
                logger.error(
                    f"Field: '{args.group_name}' not found in file: '{results_file}', " f"can't visualize results."
                )
                continue

            group = f[args.group_name]
            scores = group["scores"][:]
            skeletons = group["skeletons"][:]

        visualizer.export_to_images(video_name=video_name, results_scores=scores, results_skel=skeletons)

    # cleanup
    shutil.rmtree(args.temp_dir)


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("dataset_path", type=str)
    parser.add_argument("--results_file", type=str, help="Path of the results file, or use default locations")
    parser.add_argument("--group_name", type=str, help="Name of the H5 group in the results file, or use default")
    parser.add_argument(
        "--video_names",
        type=str,
        nargs="+",
        help="Only analyze a subset of videos. If not set, will analyze all videos in dataset_path.",
    )
    parser.add_argument("--temp_dir", type=str, help="Where to store temporary intermediate results")
    parser.add_argument("--work_dir", type=str, help="Root folder for all experiments")
    parser.add_argument(
        "--no_draw_original",
        dest="draw_original",
        action="store_false",
        help="If set, don't draw the original skeleton on the result image",
    )
    add_config_argument(parser)
    args = parser.parse_args()

    visualize(**vars(args))


if __name__ == "__main__":
    main()



"""
Utility functions to deal with eigenworms
"""

import numpy as np


def load_eigenworms_matrix(eigenworms_matrix_path: str) -> np.ndarray:
    """
    Load eigenworms matrix into numpy array from csv file

    :param eigenworms_matrix_path: path of the csv file
    :return: numpy array of the eigenworms matrix
    """
    return (
        np.loadtxt(eigenworms_matrix_path, delimiter=",").astype(float) if eigenworms_matrix_path is not None else None
    )


def theta_to_modes(theta: np.ndarray, eigenworms_matrix: np.ndarray) -> np.ndarray:
    """
    Convert angles to modes with an eigenworms matrix. We subtract the mean angle before converting.

    :param theta: angle vector, numpy array of shape (N,)
    :param eigenworms_matrix:
    :return: the modes corresponding to the angles
    """
    return eigenworms_matrix.dot(theta - np.mean(theta))


def modes_to_theta(modes: np.ndarray, eigenworms_matrix: np.ndarray) -> np.ndarray:
    """
    Convert modes to angles with an eigenworms matrix

    :param modes:
    :param eigenworms_matrix:
    :return: the angles corresponding to the modes
    """
    return modes.dot(eigenworms_matrix[:, : len(modes)].T)



#!/usr/bin/env python

"""
Entry point for all WormPose commands
"""

import importlib
import sys
import argparse

from wormpose import __version__


def main():

    entry_points = {
        "datagen": "generate_training_data",
        "train": "train_model",
        "predict": "predict_dataset",
        "postprocess": "postprocess_results",
        "viz": "visualize_results",
        "calibrate": "calibrate_dataset",
        "evaluate": "evaluate_model",
        "export": "export_results",
    }

    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "command", type=str, choices=entry_points.keys(), help="Choose which command to run",
    )
    parser.add_argument("-v", "--version", action="version", version=__version__)
    parser.parse_known_args()

    command = sys.argv.pop(1)
    module_name = entry_points[command]
    module = importlib.import_module(".".join(["wormpose", "commands", module_name]))
    module.main()


if __name__ == "__main__":
    main()



"""
Implementation of BaseFeaturesDataset to load Tierpsy tracker features
"""

import glob
import json
import os

import h5py
import numpy as np
from xml.dom import minidom

from wormpose.dataset.base_dataset import BaseFeaturesDataset


def get_frame_rate(f):
    mask_attrs = f["mask"].attrs
    if "fps" in mask_attrs:
        frame_rate = mask_attrs["fps"]
    elif "expected_fps" in mask_attrs:
        frame_rate = mask_attrs["expected_fps"]
    else:
        frame_rate = 1
    return frame_rate


def get_ventral_side(f):
    # Trying to get the ventral side attribute (optional)
    try:
        ventral_side = json.loads(f["experiment_info"][()].decode("utf-8"))["ventral_side"]
    except:
        ventral_side = None
    return ventral_side


def get_stage_position(f):
    len_frames = len(f["mask"])
    try:
        stage_position_pix = f["stage_position_pix"][:]
    except:
        stage_position_pix = np.zeros((len_frames, 2), dtype=float)
    return stage_position_pix


def get_ratio_microns_pixels(f):
    try:
        xml_info = f["xml_info"][()]
        xml = minidom.parseString(xml_info)
        microns = xml.getElementsByTagName("microns")[0].getElementsByTagName("x")[0].firstChild.nodeValue
        pixels = xml.getElementsByTagName("pixels")[0].getElementsByTagName("x")[0].firstChild.nodeValue
        ratio_microns_pixel = abs(float(microns)) / abs(float(pixels))
    except:
        ratio_microns_pixel = 1.0

    return ratio_microns_pixel


def _match_indexes(x, y):
    sorted_index = np.searchsorted(x, y)
    yindex = np.take(np.arange(len(x)), sorted_index, mode="clip")
    mask = x[yindex] != y
    result = np.ma.array(yindex, mask=mask)
    return result


def _get_width_measurement(name, ratio_microns_pixel, features_timeseries):
    feature_names = features_timeseries.dtype.names
    measurement_name = list(filter(lambda x: name in x and "width" in x, feature_names))[0]
    measurement = features_timeseries[measurement_name] * ratio_microns_pixel
    return measurement


def get_skeletons_timestamp(features_f, skeletons, features_timestamp):
    if "trajectories_data" in features_f:

        trajectories_data = features_f.get("trajectories_data")
        dt = [("skeleton_id", int)]
        with trajectories_data.astype(dt):
            skeletons_id = trajectories_data["skeleton_id"][:]

        skeletons_timestamp = np.zeros(len(skeletons), dtype=int)
        for i in range(len(skeletons)):
            skeletons_timestamp[i] = np.where(skeletons_id == i)[0][0]
    else:
        skeletons_timestamp = features_timestamp

    return skeletons_timestamp


def _resample(series, cur_timestamp, new_timestamp):
    len_new = len(new_timestamp)
    resampled_series = np.full((len_new,) + series.shape[1:], np.nan, dtype=series.dtype)

    matched_indexes = _match_indexes(cur_timestamp, new_timestamp)

    for index in range(len_new):

        if not matched_indexes.mask[index]:
            resampled_series[index] = series[matched_indexes[index]]

    return resampled_series


def get_features_filename(root_dir: str, name: str):
    """
    The features filename has different formats:
    ex: videoname_features.hdf5 or sometimes videoname_featuresN.hdf5
    """
    return glob.glob(os.path.join(root_dir, name, name + "*features*.hdf5"))[0]


def get_frames_filename(root_dir: str, name: str):
    return os.path.join(root_dir, name, name + ".hdf5")


def _read_features(root_dir, name):
    h5file = get_frames_filename(root_dir, name)
    h5featurefile = get_features_filename(root_dir, name)

    with h5py.File(h5file, "r") as f:
        frames_timestamp = f["timestamp"]["raw"][:]
        frame_rate = get_frame_rate(f)
        ventral_side = get_ventral_side(f)
        stage_position_pix = get_stage_position(f)
        ratio_microns_pixel = get_ratio_microns_pixels(f)

    with h5py.File(h5featurefile, "r") as f:
        skeletons = f["coordinates"]["skeletons"][:]
        features_timeseries = get_features_timeseries(f)
        features_timestamp = features_timeseries["timestamp"].astype(int)
        skeletons_timestamp = get_skeletons_timestamp(f, skeletons, features_timestamp)

        head_width = _get_width_measurement("head", ratio_microns_pixel, features_timeseries)
        midbody_width = _get_width_measurement("midbody", ratio_microns_pixel, features_timeseries)
        tail_width = _get_width_measurement("tail", ratio_microns_pixel, features_timeseries)
        measurements = np.stack([head_width, midbody_width, tail_width], axis=1)

    measurements = _resample(measurements, cur_timestamp=features_timestamp, new_timestamp=frames_timestamp)
    skeletons = _resample(skeletons, cur_timestamp=skeletons_timestamp, new_timestamp=frames_timestamp)

    # convert skeletons coordinates from microns to pixels
    skeletons = skeletons * ratio_microns_pixel - stage_position_pix[:, np.newaxis, :]

    head_width = measurements[:, 0]
    midbody_width = measurements[:, 1]
    tail_width = measurements[:, 2]

    return {
        "skeletons": skeletons,
        "head_width": head_width,
        "midbody_width": midbody_width,
        "tail_width": tail_width,
        "frame_rate": frame_rate,
        "ventral_side": ventral_side,
        "timestamp": frames_timestamp,
    }


def get_features_timeseries(f):
    features_timeseries = f[list(filter(lambda x: "timeseries" in x, f))[0]][:]
    # only use one worm (smallest index)
    worm_indexes = features_timeseries["worm_index"]
    worm_index = np.min(worm_indexes)
    cur_worm = np.where(worm_indexes == worm_index)[0]
    features_timeseries = features_timeseries[cur_worm]
    return features_timeseries


class FeaturesDataset(BaseFeaturesDataset):
    def __init__(self, dataset_path, video_names):
        self._features = {}
        for video_name in video_names:
            self._features[video_name] = _read_features(dataset_path, video_name)

    def get_features(self, video_name):
        return self._features[video_name]



"""
Implementation of BaseFramesDataset to load Tierpsy tracker frames
"""

import os

import h5py

from wormpose import BaseFramesDataset


class FramesDataset(BaseFramesDataset):
    class FramesReader(object):
        def __init__(self, h5_filename: str):
            self.h5_filename = h5_filename

        def __enter__(self):
            self.f = h5py.File(self.h5_filename, "r")
            return self.f["mask"]

        def __exit__(self, exc_type, exc_value, traceback):
            self.f.close()

    def __init__(self, dataset_path: str):

        if not os.path.isdir(dataset_path):
            raise FileNotFoundError(f"'{dataset_path}' needs to be a directory containing Tierpsy videos.")

        self.videos_paths = {}

        for video_name in os.listdir(dataset_path):
            if os.path.isdir(os.path.join(dataset_path, video_name)):
                video_path = os.path.join(dataset_path, video_name, video_name + ".hdf5")
                if os.path.exists(video_path):
                    self.videos_paths[video_name] = video_path

        if len(self.videos_paths) == 0:
            raise FileNotFoundError(f"Couldn't find Tierpsy videos in '{dataset_path}'.")

    def video_names(self):
        return list(self.videos_paths.keys())

    def open(self, video_name: str):
        return self.FramesReader(self.videos_paths[video_name])



"""
Exports WormPose results (skeleton only) in WCON format

It requires the original WCON file from Zenodo as an input so that we can write the same metadata
"""
import json
import os
import zipfile

import h5py
import numpy as np

import wormpose
from wormpose import BaseResultsExporter
from wormpose.dataset.loaders.tierpsy.features_dataset import (
    get_frames_filename,
    get_stage_position,
    get_ratio_microns_pixels,
)


class ResultsExporter(BaseResultsExporter):
    """
    Exports WormPose results to WCON
    If there is an already existing WCON, load it to copy the metadata field
    """

    def __init__(self, dataset_path: str):
        self.dataset_path = dataset_path

    def _find_wcon_file(self, video_name) -> str:
        """
        Will unzip the wcon file if necessary
        """
        wcon_filename = os.path.join(self.dataset_path, video_name, video_name + ".wcon")
        if os.path.exists(wcon_filename):
            return wcon_filename

        wcon_zipfilename = os.path.join(self.dataset_path, video_name, video_name + ".wcon.zip")
        if os.path.exists(wcon_zipfilename):
            with zipfile.ZipFile(wcon_zipfilename, "r") as zip_ref:
                zip_ref.extractall(os.path.join(self.dataset_path, video_name))
            return wcon_filename

        raise FileNotFoundError(f"Missing the original WCON file for {video_name}")

    def export(self, video_name: str, **kwargs):

        timestamp = kwargs["dataset"].features_dataset[video_name].timestamp
        results_skeletons: np.ndarray = kwargs["results_skeletons"]
        out_dir: str = kwargs["out_dir"]
        compress: bool = kwargs["compress"] if "compress" in kwargs else True

        try:
            wcon_filename = self._find_wcon_file(video_name)
            with open(wcon_filename, "r") as f:
                metadata = json.load(f)["metadata"]
        except Exception:
            metadata = {}
        wcon_content = {"metadata": metadata, "units": {}, "data": [{"id": "1"}]}

        h5file = get_frames_filename(self.dataset_path, video_name)
        with h5py.File(h5file, "r") as f:
            stage_position_pix = get_stage_position(f)
            ratio_microns_pixel = get_ratio_microns_pixels(f)
            frames_timestamp = f["timestamp"]["time"][:]

        wcon_content["metadata"]["software"] = {
            "name": "WormPose (https://github.com/iteal/wormpose)",
            "version": wormpose.__version__,
        }
        wcon_content["units"] = {"x": "micrometers", "y": "micrometers", "t": "seconds"}

        none_results = [None] * len(results_skeletons[0])

        data = wcon_content["data"][0]
        t_series = []
        x_series = []
        y_series = []

        for cur_time, skel in enumerate(results_skeletons):
            frame_index = np.where(timestamp == cur_time)[0]
            if len(frame_index) == 0:
                continue
            result_skel_microns = (skel + stage_position_pix[frame_index[0]]) / ratio_microns_pixel
            if np.any(np.isnan(result_skel_microns)):
                x_series.append(none_results)
                y_series.append(none_results)
            else:
                x_series.append(result_skel_microns[:, 0].tolist())
                y_series.append(result_skel_microns[:, 1].tolist())
            t_series.append(frames_timestamp[frame_index[0]])

        data["x"] = x_series
        data["y"] = y_series
        data["t"] = t_series

        exported_wcon = os.path.join(out_dir, video_name + ".wcon")
        with open(exported_wcon, "w") as f:
            json.dump(wcon_content, f)

        if compress:
            with zipfile.ZipFile(exported_wcon + ".zip", "w", compression=zipfile.ZIP_DEFLATED) as zipf:
                zipf.write(exported_wcon, arcname=os.path.basename(exported_wcon))

            os.remove(exported_wcon)



"""
Definition of the network model, a ResNet with three blocks
"""

import tensorflow as tf


def build_model(input_shape, out_dim, activation=tf.keras.layers.LeakyReLU):
    filters = 32

    def middle_stack(x, activation):
        x = _res_block(x, filters=filters, num_blocks=3, strides=2, name="res32", activation=activation,)
        x = _res_block(x, filters=filters * 2, num_blocks=3, strides=2, name="res64", activation=activation,)
        x = _res_block(x, filters=filters * 4, num_blocks=3, strides=2, name="res128", activation=activation,)
        return x

    return _build_model(input_shape, out_dim, middle_stack, filters, activation)


def _build_model(input_shape, out_dim, middle_stack, filters, activation):
    inputs = tf.keras.Input(shape=(input_shape[0], input_shape[1], 1))

    x = tf.keras.layers.Conv2D(filters=filters, kernel_size=7, strides=2, padding="same")(inputs)
    x = tf.keras.layers.MaxPooling2D(pool_size=2, strides=2, padding="same")(x)

    x = middle_stack(x, activation=activation)

    x = tf.keras.layers.BatchNormalization()(x)
    x = activation()(x)
    x = tf.keras.layers.GlobalAveragePooling2D()(x)
    x = tf.keras.layers.Dense(units=out_dim, name="pred")(x)

    model = tf.keras.Model(inputs=inputs, outputs=x)
    return model


def _basic_block(x, filters, strides, name, activation):
    x = tf.keras.layers.BatchNormalization(name=name + "_bn_0")(x)
    x = activation(name=name + "_act_0")(x)
    shortcut = tf.keras.layers.Conv2D(
        filters=filters, kernel_size=1, padding="valid", strides=strides, name=name + "_conv_0",
    )(x)
    x = tf.keras.layers.Conv2D(filters=filters, kernel_size=3, padding="same", strides=strides, name=name + "_conv_1",)(
        x
    )
    x = tf.keras.layers.BatchNormalization(name=name + "_bn_1")(x)
    x = activation(name=name + "_act_1")(x)
    x = tf.keras.layers.Conv2D(filters=filters, kernel_size=3, padding="same", strides=1, name=name + "_conv_2")(x)

    x = tf.keras.layers.Add(name=name + "_add")([x, shortcut])

    return x


def _res_block(x, filters, num_blocks, strides, name, activation):
    strides = [strides] + [1] * (num_blocks - 1)
    for index, stride in enumerate(strides):
        x = _basic_block(x=x, filters=filters, strides=stride, name=f"{name}_block_{index}", activation=activation,)
    return x



"""
Functions that deal with the image similarity
"""

import cv2
import numpy as np


def calculate_similarity(source_image: np.ndarray, template_image: np.ndarray):
    """
    Calculate how similar a template image is to a source image

    :param source_image: image to be compared to
    :param template_image: image to compare to the source_image
    :return: similarity score and position of the best match
    """

    # calculate cross correlation between a source and a template image
    score_map = cv2.matchTemplate(source_image, template_image, method=cv2.TM_CCOEFF_NORMED)

    # find the maximum value (and its location) of the cross correlation map
    _, score, _, score_loc = cv2.minMaxLoc(score_map)

    # we take the absolute value of the cross correlation as our image similarity
    similarity = abs(score)

    return similarity, score_loc


def fit_bounding_box_to_worm(worm_image: np.ndarray, background_color: int, padding: int = 2):
    """
    Calculates the bounding box of a worm in a real processed or a synthetic image:
    it should contain a worm object in the center and have a uniform background color
    The bounding box will be a little bigger than the worm (padding param)

    :param worm_image: image containing a worm
    :param background_color: the color of the background (should be uniform)
    :param padding: extra padding that will be added around the worm bounding box
    :return: coordinates left, right, bottom, top of the bounding box of the worm
    """
    where_mask = np.where(worm_image != background_color)
    if len(where_mask[0]) > 0 and len(where_mask[1]) > 0:
        left, bottom = np.clip(np.min(where_mask, axis=1) - padding, a_min=0, a_max=worm_image.shape)
        right, top = np.clip(np.max(where_mask, axis=1) + 1 + padding, a_min=0, a_max=worm_image.shape)
    else:
        left, right, bottom, top = 0, -1, 0, -1

    return left, right, bottom, top



"""
Wrapper to save the training data to different file formats
"""


class GenericFileWriter(object):
    """
    Write data to different file formats depending on the open_file and write_file functions
    """

    def __init__(self, open_file=None, write_file=None):
        self.open_file = open_file
        self.write_file = write_file

    def __enter__(self):
        self.f = self.open_file()
        self.f.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.f.__exit__(exc_type, exc_val, exc_tb)

    def write(self, data):
        self.write_file(self.f, data)



"""
Module containing everything related to the image similarity,
including the systems to distribute the image score calculation to several processes when analyzing a video
"""

from wormpose.images.scoring.scoring_data_manager import BaseScoringDataManager, ScoringDataManager
from wormpose.images.scoring.results_scoring import ResultsScoring



#!/usr/bin/env python

"""
Predicts videos using a trained model
"""

import logging
import multiprocessing as mp
import os
import random
import shutil
import tempfile
from argparse import Namespace
from functools import partial
from typing import Tuple

import numpy as np
import tensorflow as tf

from wormpose.commands import _log_parameters
from wormpose.commands.utils.results_saver import ResultsSaver
from wormpose.commands.utils.time_sampling import resample_results
from wormpose.config import default_paths
from wormpose.config.default_paths import RESULTS_FILENAME, CONFIG_FILENAME
from wormpose.config.experiment_config import load_config, add_config_argument
from wormpose.dataset.features import Features
from wormpose.dataset.loader import get_dataset_name
from wormpose.dataset.loader import load_dataset
from wormpose.dataset.loaders.resizer import ResizeOptions
from wormpose.images.scoring import BaseScoringDataManager, ScoringDataManager, ResultsScoring
from wormpose.machine_learning.best_models_saver import BestModels
from wormpose.machine_learning.predict_data_generator import PredictDataGenerator
from wormpose.pose.centerline import skeletons_to_angles
from wormpose.pose.headtail_resolution import resolve_head_tail
from wormpose.pose.results_datatypes import (
    ShuffledResults,
    OriginalResults,
    BaseResults,
)

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
tf.get_logger().setLevel(logging.INFO)


def _make_tf_dataset(data_generator, batch_size: int, image_shape):
    def run(video_name):
        data_gen = partial(data_generator.run, video_name=video_name)
        tf_dset = tf.data.Dataset.from_generator(
            data_gen, tf.float32, tf.TensorShape([batch_size, image_shape[0], image_shape[1], 1]),
        )
        return tf_dset

    return run


def _can_resolve_results(shuffled_results: ShuffledResults, score_threshold: float, video_name: str) -> bool:
    scores = shuffled_results.scores
    if np.all(np.isnan(scores)):
        logger.error(f"Calculated scores are all invalid, stopping analysis for {video_name}")
        return False

    if np.max(scores) < score_threshold:
        logger.error(
            f"There is not one frame where the error metric is above the threshold {score_threshold} "
            f"in the whole video {video_name}, stopping analysis. Maybe the model didn't train properly."
        )
        return False
    return True


class _Predictor(object):
    def __init__(self, results_scoring: ResultsScoring, keras_model):
        self.keras_model = keras_model
        self.results_scoring = results_scoring

    def __call__(
        self, num_frames: int, input_frames, scoring_data_manager: BaseScoringDataManager, features: Features,
    ) -> Tuple[OriginalResults, ShuffledResults]:
        # run all frames through the neural network to get a result theta without head/tail decision
        network_predictions = self.keras_model.predict(input_frames)[:num_frames]
        logger.info(f"Predicted {len(network_predictions)} frames")

        shuffled_results = ShuffledResults(random_theta=network_predictions)

        original_results = OriginalResults(
            theta=skeletons_to_angles(features.skeletons, theta_dims=network_predictions.shape[1]),
            skeletons=features.skeletons,
            scores=None,
        )

        # calculate image similarity for each frame, for the two solutions
        self.results_scoring(results=shuffled_results, scoring_data_manager=scoring_data_manager)

        avg_score = np.max(shuffled_results.scores, axis=1).mean()
        logger.info(f"Calculated image similarity, average: {avg_score:.4f}")

        resample_results(shuffled_results, features.timestamp)
        resample_results(original_results, features.timestamp)

        return original_results, shuffled_results


def _apply_resize_factor(results: BaseResults, resize_factor: float):
    results.skeletons /= resize_factor


def _parse_arguments(dataset_path: str, kwargs: dict):
    if kwargs.get("work_dir") is None:
        kwargs["work_dir"] = default_paths.WORK_DIR
    if kwargs.get("num_process") is None:
        kwargs["num_process"] = os.cpu_count()
    if kwargs.get("temp_dir") is None:
        kwargs["temp_dir"] = tempfile.gettempdir()
    if kwargs.get("batch_size") is None:
        kwargs["batch_size"] = 512
    if kwargs.get("score_threshold") is None:
        kwargs["score_threshold"] = 0.7
    if kwargs.get("video_names") is None:
        kwargs["video_names"] = None
    if kwargs.get("random_seed") is None:
        kwargs["random_seed"] = None

    kwargs["temp_dir"] = tempfile.mkdtemp(dir=kwargs["temp_dir"])

    dataset_name = get_dataset_name(dataset_path)
    kwargs["experiment_dir"] = os.path.join(kwargs["work_dir"], dataset_name)

    if kwargs.get("model_path") is None:
        default_models_dir = os.path.join(kwargs["experiment_dir"], default_paths.MODELS_DIRS)
        kwargs["model_path"] = BestModels(default_models_dir).best_model_path
    if kwargs.get("config") is None:
        kwargs["config"] = os.path.join(kwargs["experiment_dir"], CONFIG_FILENAME)

    _log_parameters(logger.info, {"dataset_path": dataset_path})
    _log_parameters(logger.info, kwargs)

    return Namespace(**kwargs)


def predict(dataset_path: str, **kwargs):
    """
    Use a trained model to predict the centerlines of worm for videos in a dataset

    :param dataset_path: Root path of the dataset containing videos of worm
    """
    args = _parse_arguments(dataset_path, kwargs)

    mp.set_start_method("spawn", force=True)

    if args.random_seed is not None:
        os.environ["TF_DETERMINISTIC_OPS"] = "1"
        random.seed(args.random_seed)
        np.random.seed(args.random_seed)
        tf.random.set_seed(args.random_seed)

    results_root_dir = os.path.join(args.experiment_dir, default_paths.RESULTS_DIR)
    os.makedirs(results_root_dir, exist_ok=True)

    config = load_config(args.config)

    dataset = load_dataset(
        dataset_loader=config.dataset_loader,
        dataset_path=dataset_path,
        selected_video_names=args.video_names,
        resize_options=ResizeOptions(resize_factor=config.resize_factor),
    )

    keras_model = tf.keras.models.load_model(args.model_path, compile=False)

    results_saver = ResultsSaver(
        temp_dir=args.temp_dir, results_root_dir=results_root_dir, results_filename=RESULTS_FILENAME
    )

    tf_dataset_maker = _make_tf_dataset(
        data_generator=PredictDataGenerator(
            dataset=dataset,
            num_process=args.num_process,
            temp_dir=args.temp_dir,
            image_shape=config.image_shape,
            batch_size=args.batch_size,
        ),
        batch_size=args.batch_size,
        image_shape=config.image_shape,
    )

    results_scoring = ResultsScoring(
        frame_preprocessing=dataset.frame_preprocessing,
        num_process=args.num_process,
        temp_dir=args.temp_dir,
        image_shape=config.image_shape,
    )
    predictor = _Predictor(results_scoring=results_scoring, keras_model=keras_model)

    for video_name in dataset.video_names:
        logger.info(f'Processing video: "{video_name}"')
        features = dataset.features_dataset[video_name]

        template_indexes = features.labelled_indexes
        if len(template_indexes) == 0:
            logger.error(
                f"Can't calculate image metric, there is no labelled frame in the video to use as a template, "
                f"stopping analysis for {video_name}."
            )
            continue

        original_results, shuffled_results = predictor(
            input_frames=tf_dataset_maker(video_name),
            num_frames=dataset.num_frames(video_name),
            features=features,
            scoring_data_manager=ScoringDataManager(
                video_name=video_name, frames_dataset=dataset.frames_dataset, features=features,
            ),
        )

        results = {"original": original_results, "unaligned": shuffled_results}
        if _can_resolve_results(shuffled_results, video_name=video_name, score_threshold=args.score_threshold,):
            final_results = resolve_head_tail(
                shuffled_results=shuffled_results,
                original_results=original_results,
                frame_rate=features.frame_rate,
                score_threshold=args.score_threshold,
            )
            results["resolved"] = final_results
            _apply_resize_factor(results["resolved"], config.resize_factor)

        _apply_resize_factor(results["unaligned"], config.resize_factor)

        results_saver.save(results=results, video_name=video_name)

    # cleanup
    shutil.rmtree(args.temp_dir)


def main():
    import argparse

    parser = argparse.ArgumentParser()

    # model infos
    parser.add_argument(
        "--model_path", type=str, help="Load model from this path, or use best model from work_dir.",
    )
    parser.add_argument("--batch_size", type=int)

    # inputs
    parser.add_argument("dataset_path", type=str)
    parser.add_argument(
        "--video_names",
        type=str,
        nargs="+",
        help="Only analyze a subset of videos. If not set, will analyze all videos in dataset_path.",
    )
    add_config_argument(parser)
    parser.add_argument("--temp_dir", type=str, help="Where to store temporary intermediate results")
    parser.add_argument("--work_dir", type=str, help="Root folder for all experiments")
    # multiprocessing params
    parser.add_argument("--num_process", type=int, help="How many worker processes")
    # parameters of results processing
    parser.add_argument(
        "--score_threshold",
        type=float,
        help="Image metric score threshold : discard results scoring lower than this value."
        " Fine tune this value using the script calibrate_dataset.py",
    )
    parser.add_argument("--random_seed", type=int, help="Optional random seed for deterministic results")
    args = parser.parse_args()

    predict(**vars(args))


if __name__ == "__main__":
    main()



"""
Save results to a H5 file
"""

import os
import shutil
from typing import Optional

import h5py


class ResultsSaver(object):
    def __init__(self, temp_dir: str, results_root_dir: str, results_filename: str):
        self.temp_dir = temp_dir
        self.results_root_dir = results_root_dir
        self.results_filename = results_filename

    def save(self, results: dict, video_name: str, metadata: Optional[dict] = None):
        results_folder = os.path.join(self.results_root_dir, video_name)
        if not os.path.exists(results_folder):
            os.mkdir(results_folder)

        results_temp_filepath = os.path.join(self.temp_dir, self.results_filename)
        with h5py.File(results_temp_filepath, "w") as f:

            if metadata is not None:
                for key, val in metadata.items():
                    f.attrs[key] = val

            for group_name, group_val in results.items():
                group = f.create_group(group_name)
                for name, val in group_val.__dict__.items():
                    if val is not None:
                        group.create_dataset(name, data=val)

        shutil.move(results_temp_filepath, os.path.join(results_folder, self.results_filename))



import math
import numpy as np

from wormpose.images.synthetic import BODY_SEGMENTS_DIVIDER, DRAW_TAIL_ON_TOP_PROBABILITY


def _make_calc_one_segment_coords():
    """
    Functor to calculate patches coordinates:
    From two centerline joint coordinates, and a known width, it will calculate
     the coordinates of a rectangle aligned in between the two joints positions.
    This rectangle represents a little segment along the worm body.
    """

    # preallocates some internal data structures
    skel_perp = np.empty((2,), dtype=np.float32)

    def run(joint_a, joint_b, width, dest):
        perp = [-(joint_b[1] - joint_a[1]), joint_b[0] - joint_a[0]]
        norm = math.sqrt(perp[0] * perp[0] + perp[1] * perp[1])

        skel_perp[0] = (perp[0] / norm) * (width / 2)
        skel_perp[1] = (perp[1] / norm) * (width / 2)

        dest[0] = joint_a + skel_perp
        dest[1] = joint_b + skel_perp
        dest[2] = joint_b - skel_perp
        dest[3] = joint_a - skel_perp

    return run


def make_calc_all_coords(nb_skeleton_joints: int, enable_random_augmentations: bool):
    """
    Calculates the coordinates of all the square patches along the worm body,
    for the template posture and the target posture, so we can do the affine transform later.
    The order of the coordinates list will define their drawing order

    :param nb_skeleton_joints: How many joints has the template skeleton, will be used to define how many joints
        to include in a patch. The size of a patch is defined by BODY_SEGMENTS_DIVIDER so it is consistent
        for different datasets with different number of skeleton joints.
    :param enable_random_augmentations: If True: Will randomly reverse the order of the list of patches
        depending on DRAW_TAIL_ON_TOP_PROBABILITY. If False, the patch corresponding to the head will always
        be last in the list so that it can be drawn last (head on top of tail)
    :return: tuple (list of template patches coordinates, list of target patches coordinates)
    """
    step = max(1, nb_skeleton_joints // BODY_SEGMENTS_DIVIDER)
    last = nb_skeleton_joints - step

    template_coords = np.empty((last + 2, 4, 2), dtype=np.float32)
    target_coords = np.empty((last + 2, 4, 2), dtype=np.float32)
    calc_one_segment_coords = _make_calc_one_segment_coords()

    def _update_coords(
        width, index_to_update, template_joint_a, template_joint_b, target_joint_a, target_joint_b,
    ):
        calc_one_segment_coords(
            joint_a=template_joint_a, joint_b=template_joint_b, width=width, dest=template_coords[index_to_update],
        )
        calc_one_segment_coords(
            joint_a=target_joint_a, joint_b=target_joint_b, width=width, dest=target_coords[index_to_update],
        )

    def run(template_skel, target_skel, target_worm_thickness):
        # first calculate the patches for the two extremities (beyond head and tail)
        # to avoid to cut the synthetic worm abruptly
        # they are first in the list, so that they get drawn first (will be the most in the background)

        # before head
        _update_coords(
            width=target_worm_thickness[0],
            template_joint_a=template_skel[0] + template_skel[0] - template_skel[step],
            template_joint_b=template_skel[0],
            target_joint_a=target_skel[0] + target_skel[0] - target_skel[step],
            target_joint_b=target_skel[0],
            index_to_update=0,
        )
        # after tail
        _update_coords(
            width=target_worm_thickness[last],
            template_joint_a=template_skel[last - 1 + step],
            template_joint_b=template_skel[last - 1 + step] + template_skel[last - 1 + step] - template_skel[last - 1],
            target_joint_a=target_skel[last - 1 + step],
            target_joint_b=target_skel[last - 1 + step] + target_skel[last - 1 + step] - target_skel[last - 1],
            index_to_update=1,
        )

        # decide about the order of the patches along the worm body
        if enable_random_augmentations and np.random.random() >= DRAW_TAIL_ON_TOP_PROBABILITY:
            # draw the worm starting from the head (the tail will be on top)
            joint_indexes = range(0, last)
        else:
            # draw starting from the tail (the head will be on top)
            joint_indexes = reversed(range(0, last))

        # calculate all the other patches in the body, in head to tail or tail to head order
        for index, joint_index in enumerate(joint_indexes):
            _update_coords(
                width=target_worm_thickness[joint_index],
                template_joint_a=template_skel[joint_index],
                template_joint_b=template_skel[joint_index + step],
                target_joint_a=target_skel[joint_index],
                target_joint_b=target_skel[joint_index + step],
                index_to_update=index + 2,
            )

        return template_coords, target_coords

    return run



"""
Builds the real dataset, where images from a dataset are preprocessed and then cropped to be the same size
"""

from typing import Tuple

import numpy as np

from wormpose import BaseFramePreprocessing
from wormpose.dataset.image_processing import frame_preprocessor


class RealDataset(object):
    """
    The RealDataset takes a raw image and apply the FramePreprocessing, it also ensures all the resulting images
    have the same size, by expanding the region if necessary, or by simply cropping
    """

    def __init__(
        self, frame_preprocessing: BaseFramePreprocessing, output_image_shape: Tuple[int, int],
    ):
        """

        :param frame_preprocessing: the FramePreprocessing object containing the image preprocessing logic
        :param output_image_shape: the desired output size of the images
        """

        if output_image_shape[0] % 2 == 1 or output_image_shape[1] % 2 == 1:
            raise NotImplementedError(
                f"Image width and height should be even numbers: {output_image_shape}, " f"odd numbers not supported."
            )

        self.output_image_shape = np.array((output_image_shape[0], output_image_shape[1]))
        self.frame_preprocessing = frame_preprocessing

    def process_frame(self, cur_frame: np.ndarray):
        """
        Processes one image

        :param cur_frame: Raw frame from a dataset to process
        :return: Processed version of the frame and normalized to the desired image size
        """

        processed_frame, bg_mean_color, worm_roi = frame_preprocessor.run(self.frame_preprocessing, cur_frame)

        center = (
            (worm_roi[0].start + worm_roi[0].stop) // 2,
            (worm_roi[1].start + worm_roi[1].stop) // 2,
        )

        # simple case: just crop the desired shape centered around the region of interest
        if _can_crop_simply(center, self.output_image_shape, processed_frame):
            roi_coord, skel_offset = _simple_crop(center, self.output_image_shape)

        # if we can't make a simple crop :
        # the region of interest of the worm is bigger than the result image or too much on the edge
        # let's copy what fits to a new empty image
        else:
            empty_frame = np.full(self.output_image_shape, bg_mean_color, dtype=processed_frame.dtype)
            processed_frame, roi_coord, skel_offset = _complex_crop(
                empty_frame=empty_frame,
                center=center,
                out_shape=self.output_image_shape,
                processed_frame=processed_frame,
                worm_roi=worm_roi,
            )

        roi = processed_frame[roi_coord]

        return roi, skel_offset


def _can_crop_simply(center, output_image_shape: np.ndarray, processed_frame):
    """
    We can make a simple crop if the output_image_shape with its center position set, fits in the processed_frame
    """
    return (center - output_image_shape // 2 >= 0).all() and (
        center + output_image_shape // 2 < processed_frame.shape
    ).all()


def _simple_crop(center, out_shape: np.ndarray):
    top_left = (center - out_shape / 2).astype(int)
    bottom_right = (center + out_shape / 2).astype(int)

    roi_coord = np.s_[top_left[0] : bottom_right[0], top_left[1] : bottom_right[1]]

    # skeleton is in XY coordinates (numpy is YX)
    skel_offset = (top_left[1], top_left[0])

    return roi_coord, skel_offset


def _complex_crop(empty_frame, center, out_shape: np.ndarray, processed_frame, worm_roi: Tuple[slice, slice]):
    """
    Creates a new uniform image of the shape out_shape (pixels are set to bg_mean_color)
    and set what fits of the processed frame in the center
    """
    out_height, out_width = out_shape

    roi_height, roi_width = (
        worm_roi[0].stop - worm_roi[0].start,
        worm_roi[1].stop - worm_roi[1].start,
    )
    y, x = max(0, (out_height - roi_height) // 2), max(0, (out_width - roi_width) // 2)
    zone_to_copy = np.s_[y : min(y + roi_height, out_height), x : min(x + roi_width, out_width)]

    if roi_height > out_height or roi_width > out_width:
        new_roi_height = zone_to_copy[0].stop - zone_to_copy[0].start
        new_roi_width = zone_to_copy[1].stop - zone_to_copy[1].start
        y, x = center[0] - new_roi_height // 2, center[1] - new_roi_width // 2
        worm_roi = np.s_[y : y + new_roi_height, x : x + new_roi_width]

    empty_frame[zone_to_copy] = processed_frame[worm_roi]
    roi_coord = np.s_[:out_height, :out_width]
    skel_offset = (
        worm_roi[1].start - zone_to_copy[1].start,
        worm_roi[0].start - zone_to_copy[0].start,
    )

    return empty_frame, roi_coord, skel_offset



#!/usr/bin/env python

"""
Generates the training and evaluation data from a dataset.
"""

import logging
import multiprocessing as mp
import os
import random
import shutil
import tempfile
import time
from argparse import Namespace

import numpy as np

from wormpose.commands import _log_parameters
from wormpose.config import default_paths
from wormpose.config.default_paths import (
    SYNTH_TRAIN_DATASET_NAMES,
    REAL_EVAL_DATASET_NAMES,
    CONFIG_FILENAME,
)
from wormpose.config.experiment_config import save_config, ExperimentConfig
from wormpose.dataset.loader import get_dataset_name
from wormpose.dataset.loader import load_dataset
from wormpose.dataset.loaders.resizer import add_resizing_arguments, ResizeOptions
from wormpose.machine_learning import eval_data_generator
from wormpose.machine_learning.synthetic_data_generator import SyntheticDataGenerator
from wormpose.machine_learning.tfrecord_file import TfrecordLabeledDataWriter
from wormpose.pose.postures_model import PosturesModel

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def _parse_arguments(kwargs: dict):
    if kwargs.get("num_process") is None:
        kwargs["num_process"] = os.cpu_count()
    if kwargs.get("temp_dir") is None:
        kwargs["temp_dir"] = tempfile.gettempdir()
    if kwargs.get("num_train_samples") is None:
        kwargs["num_train_samples"] = int(5e5)
    if kwargs.get("num_eval_samples") is None:
        kwargs["num_eval_samples"] = int(1e4)
    if kwargs.get("work_dir") is None:
        kwargs["work_dir"] = default_paths.WORK_DIR
    if kwargs.get("postures_generation") is None:
        kwargs["postures_generation"] = PosturesModel().generate
    if kwargs.get("video_names") is None:
        kwargs["video_names"] = None
    if kwargs.get("random_seed") is None:
        kwargs["random_seed"] = None
    kwargs["temp_dir"] = tempfile.mkdtemp(dir=kwargs["temp_dir"])
    kwargs["resize_options"] = ResizeOptions(**kwargs)

    _log_parameters(logger.info, kwargs)

    return Namespace(**kwargs)


def generate(dataset_loader: str, dataset_path: str, **kwargs):
    """
    Generate synthetic images (training data) and processed real images (evaluation data)
    and save them to TFrecord files using multiprocessing

    :param dataset_loader: Name of the dataset loader, for example "tierpsy"
    :param dataset_path: Root path of the dataset containing videos of worm
    """
    _log_parameters(logger.info, {"dataset_loader": dataset_loader, "dataset_path": dataset_path})
    args = _parse_arguments(kwargs)

    mp.set_start_method("spawn", force=True)

    random.seed(args.random_seed)
    np.random.seed(args.random_seed)

    # setup folders
    if not os.path.exists(args.work_dir):
        os.mkdir(args.work_dir)
    experiment_dir = os.path.join(args.work_dir, get_dataset_name(dataset_path))
    os.makedirs(experiment_dir, exist_ok=True)
    tfrecords_dataset_root = os.path.join(experiment_dir, default_paths.TRAINING_DATA_DIR)
    if os.path.exists(tfrecords_dataset_root):
        shutil.rmtree(tfrecords_dataset_root)

    dataset = load_dataset(
        dataset_loader=dataset_loader,
        dataset_path=dataset_path,
        resize_options=args.resize_options,
        selected_video_names=args.video_names,
    )

    start = time.time()
    synthetic_data_generator = SyntheticDataGenerator(
        num_process=args.num_process,
        temp_dir=args.temp_dir,
        dataset=dataset,
        postures_generation_fn=args.postures_generation,
        enable_random_augmentations=True,
        writer=TfrecordLabeledDataWriter,
        random_seed=args.random_seed,
    )
    gen = synthetic_data_generator.generate(
        num_samples=args.num_train_samples, file_pattern=os.path.join(args.temp_dir, SYNTH_TRAIN_DATASET_NAMES),
    )
    for progress in gen:
        yield progress
    yield 1.0

    theta_dims = len(next(args.postures_generation()))
    num_eval_samples = eval_data_generator.generate(
        dataset=dataset,
        num_samples=args.num_eval_samples,
        theta_dims=theta_dims,
        file_pattern=os.path.join(args.temp_dir, REAL_EVAL_DATASET_NAMES),
    )

    shutil.copytree(args.temp_dir, tfrecords_dataset_root)
    save_config(
        ExperimentConfig(
            dataset_loader=dataset_loader,
            image_shape=dataset.image_shape,
            theta_dimensions=theta_dims,
            num_train_samples=args.num_train_samples,
            num_eval_samples=num_eval_samples,
            resize_factor=args.resize_options.resize_factor,
            video_names=dataset.video_names,
        ),
        os.path.join(experiment_dir, CONFIG_FILENAME),
    )

    end = time.time()
    logger.info(f"Done generating training data in : {end - start:.1f}s")


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("dataset_loader", type=str)
    parser.add_argument("dataset_path", type=str)
    parser.add_argument(
        "--video_names",
        type=str,
        nargs="+",
        help="Only generate training data for a subset of videos. "
        "If not set, will include all videos in dataset_path.",
    )
    parser.add_argument("--num_train_samples", type=int, help="How many training samples to generate")
    parser.add_argument("--num_eval_samples", type=int, help="How many evaluation samples to generate")
    parser.add_argument("--temp_dir", type=str, help="Where to store temporary intermediate results")
    parser.add_argument("--work_dir", type=str, help="Root folder for all experiments")
    parser.add_argument("--num_process", type=int, help="How many worker processes")
    parser.add_argument("--random_seed", type=int, help="Optional random seed for deterministic results")
    add_resizing_arguments(parser)
    args = parser.parse_args()

    last_progress = None
    for progress in generate(**vars(args)):
        prog_percent = int(progress * 100)
        if prog_percent != last_progress:
            logger.info(f"Generating training data: {prog_percent}% done")
        last_progress = prog_percent


if __name__ == "__main__":
    main()



"""
Module responsible for drawing the synthetic worm image
"""

import random
from typing import Tuple

import cv2
import numpy as np

from wormpose import BaseFramePreprocessing
from wormpose.dataset.image_processing import frame_preprocessor
from wormpose.images.synthetic import (
    MAX_OFFSET_PERCENT,
    GAUSSIAN_BLUR_MIN_SIZE,
    GAUSSIAN_BLUR_MAX_SIZE_PERCENT,
    GAUSSIAN_BLUR_MAX_SIZE,
    THICKNESS_MULTIPLIER_OFFSET,
    THICKNESS_MULTIPLIER,
    WORM_LENGTH_MULTIPLIER_OFFSET,
    BLUR_PROBABILITY,
)
from wormpose.images.synthetic._coord_transform import make_calc_all_coords
from wormpose.images.synthetic._helpers import WormOutlineMask, PatchDrawing, TargetSkeletonCalculator
from wormpose.pose.centerline import get_joint_indexes


def _make_synth_worm_measurements(nb_skeleton_joints: int, measurements):
    """
    Transforms the original measurements to make them usable by the synthetic dataset generator
    The width of the worm is interpolated between the known joints (head, midbody and tail)
    """
    head_width = np.nanmean(measurements["head_width"])
    midbody_width = np.nanmean(measurements["midbody_width"])
    tail_width = np.nanmean(measurements["tail_width"])

    head_joint, mid_body_joint, tail_body_joint = get_joint_indexes(nb_skeleton_joints)
    worm_thickness = np.empty(nb_skeleton_joints, np.float32)
    worm_thickness[:head_joint] = head_width
    worm_thickness[head_joint:mid_body_joint] = np.linspace(head_width, midbody_width, mid_body_joint - head_joint)
    worm_thickness[mid_body_joint:tail_body_joint] = np.linspace(
        midbody_width, tail_width, tail_body_joint - mid_body_joint
    )
    worm_thickness[tail_body_joint:] = tail_width

    worm_length = np.nanmean(measurements["worm_length"])

    return worm_length, worm_thickness


class SyntheticDataset(object):
    """
    Class responsible for generating synthetic images
    """

    def __init__(
        self, frame_preprocessing: BaseFramePreprocessing, enable_random_augmentations: bool, output_image_shape,
    ):
        """

        :param frame_preprocessing: the FramePreprocessing object containing the image preprocessing logic
        :param enable_random_augmentations: Adds augmentation to the syntthetic images if set to True:
            translation, scale, blur
        :param output_image_shape: Desired size of the synthetic images
        """

        self.output_image_shape = (output_image_shape[0], output_image_shape[1])
        self.enable_random_augmentations = enable_random_augmentations
        self.frame_preprocessing = frame_preprocessing

        self._max_offset = (
            int(self.output_image_shape[0] * MAX_OFFSET_PERCENT),
            int(self.output_image_shape[1] * MAX_OFFSET_PERCENT),
        )
        self._nb_skeleton_joints = None
        self._theta_dims = None
        self._calc_all_coords = None
        self._synth_worm_measurements_cache = {}
        self._worm_outline_mask = WormOutlineMask(self.output_image_shape)
        self._frames_infos_cache = {}
        self._patch_drawing = PatchDrawing(output_image_shape=self.output_image_shape)
        self._target_skeleton_calculator = None
        self._target_skeleton = None
        self._gaussian_blur_filter_sizes = None

        # blurring image when random augmentations apply :
        # blur kernel size is chosen between GAUSSIAN_BLUR_MIN_SIZE and GAUSSIAN_BLUR_MAX_SIZE
        # (or a GAUSSIAN_BLUR_MAX_SIZE_PERCENT of the image size if smaller than GAUSSIAN_BLUR_MAX_SIZE)
        if self.enable_random_augmentations:
            self._gaussian_blur_filter_sizes = np.arange(
                GAUSSIAN_BLUR_MIN_SIZE,
                min(GAUSSIAN_BLUR_MAX_SIZE + 1, int(GAUSSIAN_BLUR_MAX_SIZE_PERCENT * min(output_image_shape))),
                2,
            )

        # pre allocate numpy arrays for calculations
        self._out_image = np.empty(self.output_image_shape, dtype=np.float32)

    def _get_recentered_frame(self, template_frame, template_skeleton):

        frame_id = hash(template_frame.tostring())

        if frame_id in self._frames_infos_cache:
            worm_roi, bg_mean_color = self._frames_infos_cache[frame_id]
        else:
            frame, bg_mean_color, worm_roi = frame_preprocessor.run(self.frame_preprocessing, template_frame)
            self._frames_infos_cache[frame_id] = (worm_roi, bg_mean_color)

        template_skeletons_recentered = template_skeleton - (worm_roi[1].start, worm_roi[0].start,)
        return template_frame[worm_roi], bg_mean_color, template_skeletons_recentered

    def _get_synth_measurements(self, worm_measurements):

        measurements_id = id(worm_measurements)

        if measurements_id not in self._synth_worm_measurements_cache:
            self._synth_worm_measurements_cache[measurements_id] = _make_synth_worm_measurements(
                self._nb_skeleton_joints, worm_measurements
            )

        return self._synth_worm_measurements_cache[measurements_id]

    def generate(
        self,
        theta: np.ndarray,
        template_frame: np.ndarray,
        template_skeleton: np.ndarray,
        template_measurements: np.ndarray,
        out_image: np.ndarray,
    ) -> Tuple[int, np.ndarray]:
        """
        Generates a synthetic image

        :param theta: desired posture (angles)
        :param template_frame: reference image
        :param template_skeleton: skeleton of the reference image
        :param template_measurements: measurements of the reference image (worm width and length)
        :param out_image: canvas to draw the synthetic image
        :return: background color and the synthetic worm skeleton coordinates
        """

        if out_image.dtype != np.uint8:
            raise NotImplementedError("Only uint8 images supported.")

        if self._nb_skeleton_joints is None:
            self._nb_skeleton_joints = len(template_skeleton)
            self._target_skeleton = np.empty_like(template_skeleton)
        if len(template_skeleton) != self._nb_skeleton_joints:
            raise NotImplementedError("Can't change the amount of skeleton joints.")

        if self._theta_dims is None:
            self._theta_dims = len(theta)
        if len(theta) != self._theta_dims:
            raise NotImplementedError("Can't change theta dimensions.")

        if self._target_skeleton_calculator is None:
            self._target_skeleton_calculator = TargetSkeletonCalculator(
                theta_dims=self._theta_dims,
                nb_skeletons_joints=self._nb_skeleton_joints,
                output_image_shape=self.output_image_shape,
            )
        if self._calc_all_coords is None:
            self._calc_all_coords = make_calc_all_coords(
                nb_skeleton_joints=self._nb_skeleton_joints,
                enable_random_augmentations=self.enable_random_augmentations,
            )

        # get image and skeleton recentered to region of interest
        recentered_frame, cur_bg_color, template_skel = self._get_recentered_frame(template_frame, template_skeleton)

        # get worm length and worm thickness from cache
        cur_worm_length, cur_worm_thickness = self._get_synth_measurements(template_measurements)

        # increase thickness to include more background pixels when making patches
        # (with some variation for data augmentation)
        thickness_multiplier_offset = THICKNESS_MULTIPLIER_OFFSET if self.enable_random_augmentations else 0
        thickness_multiplier = np.random.uniform(
            THICKNESS_MULTIPLIER - thickness_multiplier_offset, THICKNESS_MULTIPLIER + thickness_multiplier_offset
        )
        target_worm_thickness = cur_worm_thickness * thickness_multiplier

        target_worm_length = cur_worm_length
        # vary target worm length for data augmentation
        if self.enable_random_augmentations:
            target_worm_length = cur_worm_length * np.random.uniform(
                1.0 - WORM_LENGTH_MULTIPLIER_OFFSET, 1.0 + WORM_LENGTH_MULTIPLIER_OFFSET
            )

        # calculate the target skeleton coordinates, apply optional image augmentation
        self._target_skeleton_calculator.calculate(
            theta=theta, worm_length=target_worm_length, out=self._target_skeleton
        )
        target_skel = self._augment_skeleton(self._target_skeleton)

        # reset image canvas where we will draw the synthetic worm
        self._out_image.fill(0.0)

        # draw all the body patches: affine transform from template_coords to target_coords
        all_template_coords, all_target_coords = self._calc_all_coords(
            template_skel, target_skel, target_worm_thickness
        )
        for template_coords, target_coords in zip(all_template_coords, all_target_coords):
            self._patch_drawing.draw_patch(
                template_coords=target_coords,
                target_coords=template_coords,
                recentered_frame=recentered_frame,
                out_image=self._out_image,
            )

        # apply the worm outline mask to hide some unwanted pixels from the patch drawing
        self._worm_outline_mask.apply(
            worm_thickness=target_worm_thickness / 2, output_image=self._out_image, target_skel=target_skel,
        )

        # all zero pixels become background color #FIXME this doesn't work if the worm image has zero values
        self._out_image[self._out_image == 0] = cur_bg_color

        # median blur to smooth out the connections between patches while preserving edge features
        cv2.medianBlur(self._out_image, 3, dst=self._out_image)

        # optional image augmentation: gaussian blur
        self._augment_extra_blur(self._out_image)

        # convert final image from float to uint8
        np.copyto(out_image, self._out_image, casting="unsafe")

        return cur_bg_color, target_skel.copy()

    def _augment_skeleton(self, target_skel):

        # random translation offset
        tx, ty = (
            (
                np.random.uniform(-self._max_offset[0], self._max_offset[0]),
                np.random.uniform(-self._max_offset[1], self._max_offset[1]),
            )
            if self.enable_random_augmentations
            else (0, 0)
        )
        target_skel[:, 0] += tx
        target_skel[:, 1] += ty

        return target_skel

    def _augment_extra_blur(self, out_image):

        if self.enable_random_augmentations and np.random.uniform() > 1 - BLUR_PROBABILITY:
            ksize = random.choice(self._gaussian_blur_filter_sizes)
            cv2.GaussianBlur(out_image, (ksize, ksize), 0, dst=out_image)



"""
Contains constant values for default paths and folder names
"""

# root directory for all experiments, results etc
WORK_DIR = "experiments"

# Root folder to store the trained models
MODELS_DIRS = "models"

# Root folder to store the tfrecord files of training and evaluation data
TRAINING_DATA_DIR = "training_data"

# Path of the json file to store configuration about the experiment (which dataset loader, which image size etc)
CONFIG_FILENAME = "config.json"

# Root folder of the prediction results and the image visualization
RESULTS_DIR = "results"

# Name of the file containing the prediction results
RESULTS_FILENAME = "results.h5"

# Interpolation and smoothed results in this file
POSTPROCESSED_RESULTS_FILENAME = "processed_results.h5"

# Root folder for the results of the calibrate script
CALIBRATION_RESULTS_DIR = "calibration"

# Filename pattern of the dataset files
SYNTH_TRAIN_DATASET_NAMES = "synthetic_{index}.tfrecord"
REAL_EVAL_DATASET_NAMES = "real_{index}.tfrecord"



"""
Simple FeaturesDataset implementation that reads features from a HDF5 file
"""

import os
from typing import List

import h5py
from glob import glob

from wormpose.dataset.base_dataset import BaseFeaturesDataset
from wormpose.pose.centerline import get_joint_indexes


class HDF5Features(BaseFeaturesDataset):
    def __init__(self, dataset_path: str, video_names: List[str]):

        features_file = glob(os.path.join(dataset_path, "*.h*5"))
        if len(features_file) == 0:
            raise FileNotFoundError(f"Can't load features, missing .h5 file in {dataset_path}")

        with h5py.File(features_file[0], "r") as f:
            width = f["width"][:]
            skeletons = f["skeletons"][:]
            framerate = f.attrs["framerate"]

        head_joint, mid_body_joint, tail_body_joint = get_joint_indexes(len(width[0]))

        features_dict = {
            "skeletons": skeletons,
            "head_width": width[:, head_joint],
            "midbody_width": width[:, mid_body_joint],
            "tail_width": width[:, tail_body_joint],
            "frame_rate": framerate,
        }

        self._features = {video_names[0]: features_dict}

    def get_features(self, video_name: str):
        return self._features[video_name]



"""
Contains the functions that deal with drawing the worm or some overlays
"""
from typing import Callable

import cv2
import numpy as np

from wormpose.pose.centerline import get_joint_indexes


def make_draw_worm_body(body_color: int = 255) -> Callable:
    """
    Functor to draw a filled worm body on an image
    as a serie of quadrilaterals with some circles at the extremities

    :param body_color: which color the worm body will be drawn (default 255)
    :return: function to call to draw a worm on an image
    """

    #  preallocates some internal data structures
    polygon = [[]] * 4
    orthogonal = np.empty((2,), dtype=np.float)
    vertices_todraw = np.empty((4, 2), dtype=int)
    cur_joint = np.empty(2, dtype=np.float)
    prev_joint = np.empty(2, dtype=np.float)

    def _draw_extremity(img, center, radius):
        cv2.circle(img, (int(center[0]), int(center[1])), int(radius), body_color, -1, 8)

    def run(worm_thickness: np.ndarray, img: np.ndarray, skeleton: np.ndarray):

        num_centerline_joints = len(worm_thickness)

        # draw two circles at the extremities
        _draw_extremity(img, skeleton[0], worm_thickness[0])
        _draw_extremity(
            img, skeleton[num_centerline_joints - 1], worm_thickness[num_centerline_joints - 1],
        )

        rx, ry = skeleton[:, 0], skeleton[:, 1]

        # draw convex quadrilateral polygons in the middle
        prev_boundary_0 = None
        prev_boundary_1 = None

        for i in range(2, num_centerline_joints):
            cur_joint[0] = rx[i]
            cur_joint[1] = ry[i]
            prev_joint[0] = rx[i - 2]
            prev_joint[1] = ry[i - 2]

            orthogonal[0] = cur_joint[1] - prev_joint[1]
            orthogonal[1] = -(cur_joint[0] - prev_joint[0])

            length = np.sqrt(orthogonal[0] * orthogonal[0] + orthogonal[1] * orthogonal[1])
            orthogonal[0] /= length
            orthogonal[1] /= length

            polygon[0] = cur_joint + orthogonal * worm_thickness[i]
            polygon[1] = cur_joint - orthogonal * worm_thickness[i]
            if prev_boundary_0 is None and prev_boundary_1 is None:
                polygon[2] = prev_joint - orthogonal * worm_thickness[i - 1]
                polygon[3] = prev_joint + orthogonal * worm_thickness[i - 1]
            else:
                polygon[3] = prev_boundary_0
                polygon[2] = prev_boundary_1

            prev_boundary_0 = polygon[0]
            prev_boundary_1 = polygon[1]

            vertices_todraw[0] = polygon[0]
            vertices_todraw[1] = polygon[1]
            vertices_todraw[2] = polygon[2]
            vertices_todraw[3] = polygon[3]
            cv2.fillConvexPoly(img, vertices_todraw, body_color, 8)

    return run


def draw_skeleton(image: np.ndarray, skeleton: np.ndarray, color, head_color):
    """
    Draw the worm centerline as lines and draw the worm head as a circle
    """
    if np.any(np.isnan(skeleton)):
        return

    skeleton = skeleton.astype(int)
    for skel_index in range(len(skeleton) - 1):
        cv2.line(
            image,
            (skeleton[skel_index][0], skeleton[skel_index][1]),
            (skeleton[skel_index + 1][0], skeleton[skel_index + 1][1]),
            color=color,
            thickness=1,
        )
    cv2.circle(image, (skeleton[0][0], skeleton[0][1]), 2, head_color, -1)


def draw_width_circle(image: np.ndarray, center, width: float, color):
    """
    Draw a circle on top of an image to represent the worm width
    """
    cv2.circle(
        image, center=(int(center[0]), int(center[1])), radius=int(width / 2), color=color,
    )


def draw_measurements(image: np.ndarray, skeleton: np.ndarray, measurements, color):
    """
    Draw overlays on top of an image to represent the worm measurements:
    circles to indicate the width at three points along the worm body: head, midbody and tail
    """
    if np.any(np.isnan(skeleton)):
        return

    head_joint, mid_body_joint, tail_body_joint = get_joint_indexes(len(skeleton))

    draw_width_circle(image, skeleton[head_joint], measurements["head_width"], color)
    draw_width_circle(image, skeleton[mid_body_joint], measurements["midbody_width"], color)
    draw_width_circle(image, skeleton[tail_body_joint], measurements["tail_width"], color)



"""
Handles all about the optional resizing of the images of a Dataset

It will modify the FeaturesDataset and the FramesDataset to update with the resized values.

The resize parameter can be specified as a single number (ex: 2 for upscale to twice the size, or 0.5 to downscale half the size)
or as the desired image size, in that case, the resize factor will be calculated depending on the average worm length
in the dataset.
"""

import copy
from typing import Type

import cv2
from wormpose.dataset.base_dataset import BaseFeaturesDataset, BaseFramesDataset

from wormpose.dataset.features import (
    calculate_max_average_worm_length,
    calculate_crop_window_size,
    MINIMUM_IMAGE_SIZE,
)


class ResizeOptions(object):
    """
    Class handling the possible resize options: either a scaling factor, or a set image size
    """

    def __init__(self, resize_factor: float = None, image_size: int = None, **kwargs):
        self.resize_factor = resize_factor
        self._image_size = image_size

        # validate
        if self._image_size is not None:
            if self._image_size % 2 != 0:
                raise NotImplementedError("If image size is set, it should be an even number")
            if self._image_size < MINIMUM_IMAGE_SIZE:
                raise ValueError(f"Image size must be bigger than {MINIMUM_IMAGE_SIZE}")

    def update_resize_factor(self, features_dataset):

        if self.resize_factor is not None:
            return

        if self._image_size is not None:
            worm_length = calculate_max_average_worm_length(features_dataset)
            self.resize_factor = self._image_size / worm_length
            return

        self.resize_factor = 1.0

    def get_image_shape(self, features_dataset):
        if self._image_size is not None:
            return self._image_size, self._image_size

        return calculate_crop_window_size(features_dataset)


def add_resizing_arguments(parser):
    """
    For command line arguments, adds two mutually exclusive resize options: either a scaling factor, or a set image size

    :param parser: Argparse parser
    """
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--resize_factor",
        type=float,
        help="Option to resize the images. "
        "Examples: '0.5' downscale half, '2' upscale twice, '1' no resizing (default)",
    )
    group.add_argument(
        "--image_size",
        type=int,
        help="Option to resize the images so that the average worm of the dataset "
        "fits into a square image of side: image_size",
    )


class ResizedFrames(object):
    def __init__(self, scale_factor: float, frames):
        self.frames = frames
        self.scale_factor = scale_factor

    def _resize(self, frame):
        return cv2.resize(frame, dsize=None, fx=self.scale_factor, fy=self.scale_factor)

    def __getitem__(self, key):
        if isinstance(key, slice):
            raise NotImplementedError("Slicing notation not implemented with resizer.")
        orig = self.frames[key]
        return self._resize(orig)

    def __len__(self):
        return len(self.frames)


class ResizedFramesReader(object):
    def __init__(self, frames_reader, resize_factor: float):
        self.frames_reader = frames_reader
        self.resize_factor = resize_factor

    def __enter__(self):
        self.resized_frames = ResizedFrames(self.resize_factor, self.frames_reader.__enter__())
        return self.resized_frames

    def __exit__(self, exc_type, exc_value, traceback):
        self.resized_frames = None
        self.frames_reader.__exit__(exc_type, exc_value, traceback)


def frames_dataset_resizer(frames_dataset_class: Type[BaseFramesDataset], resize_factor: float):
    """
    Modifies a FramesDataset class so that all frames are resized by resize_factor

    :param frames_dataset_class: original FramesDataset class without resizing
    :param resize_factor: by how much to rescale the images
    :return: new FramesDataset with resizing
    """
    old_open = frames_dataset_class.open

    def resizer_open(*args):
        frames_reader = old_open(*args)
        return ResizedFramesReader(frames_reader, resize_factor=resize_factor)

    frames_dataset_class.open = resizer_open
    return frames_dataset_class


def features_dataset_resizer(features_dataset_class: Type[BaseFeaturesDataset], resize_factor: float):
    """
    Modifies a FeaturesDataset class so that all features are resized by resize_factor

    :param frames_dataset_class: original FeaturesDataset class without resizing
    :param resize_factor: by how much to rescale the features
    :return: new FeaturesDataset with resizing
    """
    old_get = features_dataset_class.get_features

    def resizer_get(*args):
        features = old_get(*args)
        new_features = copy.deepcopy(features)
        new_features["skeletons"] *= resize_factor
        new_features["head_width"] *= resize_factor
        new_features["midbody_width"] *= resize_factor
        new_features["tail_width"] *= resize_factor
        return new_features

    features_dataset_class.get_features = resizer_get
    return features_dataset_class
